#!/usr/bin/env python3
import argparse
import os
import signal
import subprocess
import sys

import entomb
from entomb import (
    constants,
    listing,
    processing,
    reporting,
    utilities,
)


def main(argv):
    """Run Entomb.

    Parameters
    ----------
    argv : list
        List of arguments from the command line, including the executable.

    Returns
    -------
    int
        The exit status.

    """
    # Set handler for interrupt and termination signals.
    signal.signal(signal.SIGINT, _signal_handler)
    signal.signal(signal.SIGTERM, _signal_handler)

    # Parse the command line input.
    args = _parse_args(argv)

    # If any optional arguments conflict, print the errors and return a
    # non-zero exit status.
    argument_conflicts = _check_argument_conflicts(args)
    if argument_conflicts:
        _print_argument_conflict_errors(argument_conflicts)
        return 1

    # Get the absolute path.
    absolute_path = _get_absolute_path(args.path)

    # If the absolute path doesn't exist, print an error and return a non-zero
    # exit status.
    if not os.path.exists(absolute_path):
        _print_path_not_found_error(args.path, absolute_path)
        return 1

    # Trigger the sudo prompt if the operation can change immutability
    # attributes and the user does not currently have root priviliges.
    root_privileges_required = not any([
        args.dry_run,
        args.list_immutable,
        args.list_mutable,
        args.report,
    ])
    user_has_root_privileges = _user_has_root_privileges()
    if root_privileges_required and not user_has_root_privileges:
        _prompt_for_sudo()

    # Print the entomb header.
    print()
    print("======")
    print("Entomb")
    print("======")
    print()

    # Print the dry run message if required.
    if args.dry_run:
        print(">> This is a dry run so no changes will be made")
        print()

    # Print the include git message if required.
    if args.include_git:
        print(">> Files in .git directories are being included")
        print()

    # Print the path type header.
    path_type = _get_path_type(absolute_path)
    utilities.print_header(path_type.title())

    # Print the path.
    print(absolute_path)
    print()

    # Print the operation header.
    utilities.print_header("Operation")

    # If a report is requested, print it then return a zero exit status.
    if args.report:
        reporting.produce_report(absolute_path, include_git=args.include_git)
        return 0

    # If a list of immutable objects is requested, print it then return a zero
    # exit status.
    if args.list_immutable:
        listing.list_files(
            absolute_path,
            immutable=True,
            include_git=args.include_git,
        )
        return 0

    # If a list of mutable objects is requested, print it then return a zero
    # exit status.
    if args.list_mutable:
        listing.list_files(
            absolute_path,
            immutable=False,
            include_git=args.include_git,
        )
        return 0

    # Set the immutable attribute for everything on the path, then return a
    # zero exit status.
    processing.process_objects(
        absolute_path,
        immutable=not args.unset,
        include_git=args.include_git,
        dry_run=args.dry_run,
    )
    return 0


def run():
    """Call main() to run Entomb.

    This is the console script entry point.

    Returns
    -------
    None

    """
    # Run Entomb, exiting with a status code when finished.
    sys.exit(main(sys.argv))


def _check_argument_conflicts(args):
    """Check whether any optional arguments conflict with each other.

    Parameters
    ----------
    args : instance of Namespace
        Parsed arguments from the command line.

    Returns
    -------
    list of str
        The list of argument conflict messages.

    """
    # Match each argument with a list of the arguments compatible with it.
    compatibility = {
        constants.DRY_RUN_ARG: [
            constants.INCLUDE_GIT_ARG,
            constants.UNSET_ARG,
        ],
        constants.INCLUDE_GIT_ARG: [
            constants.DRY_RUN_ARG,
            constants.LIST_IMMUTABLE_ARG,
            constants.LIST_MUTABLE_ARG,
            constants.REPORT_ARG,
            constants.UNSET_ARG,
        ],
        constants.LIST_IMMUTABLE_ARG: [constants.INCLUDE_GIT_ARG],
        constants.LIST_MUTABLE_ARG: [constants.INCLUDE_GIT_ARG],
        constants.REPORT_ARG: [constants.INCLUDE_GIT_ARG, constants.UNSET_ARG],
        constants.UNSET_ARG: [
            constants.DRY_RUN_ARG,
            constants.INCLUDE_GIT_ARG,
        ],
    }

    # Make a list of the optional arguments supplied at the command line. The
    # arguments are represented in the Namespace object with underscores
    # raplacing hyphens, so convert between underscores and hyphens are
    # required.
    optional_args = compatibility.keys()
    optional_args_with_underscores = [
        oa[2:].replace("-", "_") for oa in optional_args
    ]
    supplied_optional_args_with_underscores = [
        oa for oa in optional_args_with_underscores if getattr(args, oa)
    ]
    supplied_optional_args = [
        "--{}".format(oa.replace("_", "-"))
        for oa in supplied_optional_args_with_underscores
    ]

    # Build a set of pairs of conflicting arguments.
    conflicting_arguments = set()
    for optional_arg in supplied_optional_args:
        other_supplied_args = [
            arg for arg in supplied_optional_args if arg != optional_arg
        ]
        compatible_args = compatibility.get(optional_arg)

        for arg in other_supplied_args:
            if arg not in compatible_args:
                argument_pair = tuple(sorted([optional_arg, arg]))
                conflicting_arguments.add(argument_pair)

    # Convert each argument pair into an error message.
    error_messages = ([
        "{} and {} cannot be passed together".format(*pair)
        for pair in conflicting_arguments
    ])

    return sorted(error_messages)


def _get_absolute_path(path):
    """Convert a path into an absolute path.

    Parameters
    ----------
    path : str
        A relative or absolute path.

    Returns
    -------
    str
        The absolute path.

    """
    if os.path.isabs(path):
        absolute_path = path
    else:
        current_working_directory = os.getcwd()
        absolute_path = os.path.join(current_working_directory, path)

    # Normalise the path to remove up-level references like "." and "..".
    absolute_path = os.path.normpath(absolute_path)

    return absolute_path


def _get_path_type(path):
    """Determine if a path is a directory, file or link.

    Parameters
    ----------
    path : str
        An absolute path.

    Returns
    -------
    str
        The type of object.

    Raises
    ------
    AssertionError
        If the path does not exist or is not an absolute path.

    """
    # Parameter check.
    assert os.path.exists(path)
    assert os.path.isabs(path)

    if os.path.islink(path):
        path_type = constants.LINK
    elif os.path.isfile(path):
        path_type = constants.FILE
    elif os.path.isdir(path):
        path_type = constants.DIRECTORY
    else:
        # For all others, use a generic term.
        path_type = "path"

    return path_type


def _parse_args(argv):
    """Parse the command line input.

    Parameters
    ----------
    argv : list of str
        List of arguments from the command line, including the executable.

    Returns
    -------
    instance of Namespace
        Parsed arguments from the command line.

    """
    parser = argparse.ArgumentParser(
        description="Manage file immutability.",
        usage="%(prog)s [options] path",
    )

    parser.add_argument("path", help="the path to operate on")
    parser.add_argument(
        constants.DRY_RUN_SHORT_ARG,
        constants.DRY_RUN_ARG,
        action="store_true",
        help="make no changes",
    )
    parser.add_argument(
        constants.INCLUDE_GIT_SHORT_ARG,
        constants.INCLUDE_GIT_ARG,
        action="store_true",
        help="include .git directories (excluded by default)",
    )
    parser.add_argument(
        constants.LIST_IMMUTABLE_ARG,
        action="store_true",
        help="list all immutable files",
    )
    parser.add_argument(
        constants.LIST_MUTABLE_ARG,
        action="store_true",
        help="list all mutable files",
    )
    parser.add_argument(
        constants.REPORT_SHORT_ARG,
        constants.REPORT_ARG,
        action="store_true",
        help="display a status report",
    )
    parser.add_argument(
        constants.UNSET_SHORT_ARG,
        constants.UNSET_ARG,
        action="store_true",
        help="unset immutability",
    )
    parser.add_argument(
        constants.VERSION_SHORT_ARG,
        constants.VERSION_ARG,
        action="version",
        version="{} {}".format(entomb.__title__, entomb.__version__),
    )

    return parser.parse_args(argv[1:])


def _print_argument_conflict_errors(conflicts):
    """Print the list of command line argument conflicts.

    Parameters
    ----------
    conflicts : list of str
        A list of argument conflict messages.

    Returns
    -------
    None

    """
    print()
    for conflict in conflicts:
        print("ERROR:", conflict)


def _print_path_not_found_error(relative_path, absolute_path):
    """Print the details of a path which can't be found.

    Parameters
    ----------
    relative_path : str
        A relative path.
    absolute_path : str
        The absolute path which the relative path expanded to.

    Returns
    -------
    None

    """
    print()
    print("ERROR: The path you gave could not be found")
    print(">> You entered this path:", relative_path)

    # If the relative path expanded to a different absolute path, display the
    # absolute path.
    if relative_path != absolute_path:
        print(">> It was expanded to this:", absolute_path)


def _prompt_for_sudo():
    """Trigger the sudo prompt for a user without root privileges.

    Returns
    -------
    None

    """
    print()
    print("** To set or unset files' immutable attributes, root")
    print("** privileges are required.")
    print()
    subprocess.run(
        ["sudo", "-v"],
        check=True,
        stderr=subprocess.STDOUT,
        stdout=subprocess.DEVNULL,
    )
    print()


def _signal_handler(signum, frame):  # pylint: disable=unused-argument
    """Print an interrupt or termination message, then exit.

    Parameters
    ----------
    signum : int
        The signal number. Either 2 (SIGINT) or 15 (SIGTERM).
    frame : frame object
        Unused.

    Returns
    -------
    None

    """
    print()
    print()
    if signum == signal.SIGINT:
        print(">> Keyboard interrupt signal received. Exiting.")
    elif signum == signal.SIGTERM:
        print(">> Termination signal received. Exiting.")
    print()
    sys.exit(1)


def _user_has_root_privileges():
    """Determine whether the user currently has root privileges.

    Returns
    -------
    bool
        Whether the user currently has root privileges.

    """
    try:
        subprocess.run(
            ["sudo", "-ln"],
            check=True,
            stderr=subprocess.STDOUT,
            stdout=subprocess.DEVNULL,
        )
        has_root_privileges = True
    except subprocess.CalledProcessError:
        has_root_privileges = False

    return has_root_privileges



import datetime
import decimal
import os
import subprocess
from contextlib import contextmanager

from entomb import (
    constants,
    exceptions,
)


def clear_line():
    """Clear the current line in the terminal.

    Returns
    -------
    None

    """
    print("\033[K", end="")


def count_file_paths(path, include_git, print_frequency=1000):
    """Count the file paths in a directory.

    Prints the file path count as it progresses.

    Parameters
    ----------
    path : str
        An absolute path.
    include_git: bool
        Whether to include git files.
    print_frequency : int, optional
        The number of file paths to count before re-printing the count. The
        default is 1000.

    Returns
    -------
    int
        The number of file paths in the directory and its subdirectories.

    """
    print("Counting file paths: 0", end="\r")

    count = 0

    # Walk the path.
    for _ in file_paths(path, include_git):
        count += 1

        # Only print the count at the specified frequency.
        if count % print_frequency == 0:
            print("Counting file paths: {:,}".format(count), end="\r")

    # Clear the final count message.
    clear_line()

    return count


def file_is_immutable(path):
    """Whether a file has the immutable attribute set.

    Parameters
    ----------
    path : str
        An absolute path.

    Returns
    -------
    bool
        True if the file's immmutable attribute is set, False if it is not.

    Raises
    ------
    AssertionError
        If the path is a directory, is a link or does not exist.
    GetAttributeError
        If the path's immutable attribute cannot be accessed.

    """
    # Parameter check.
    assert not os.path.isdir(path)
    assert not os.path.islink(path)
    assert os.path.exists(path)

    # Get the immutable flag.
    immutable_flag = _get_immutable_flag(path)

    return immutable_flag == "i"


def file_paths(path, include_git):
    """Generate paths of all files and links on the path.

    Parameters
    ----------
    path : str
        An absolute path.
    include_git: bool
        Whether to include git files.

    Yields
    ------
    str
        An absolute path.

    Raises
    ------
    AssertionError
        If the path does not exist.

    """
    # Parameter check.
    # Note that this assert statement appears to only raise an AssertionError
    # when the generator is iterated, not when it is created.
    assert os.path.exists(path)

    # Yield the path if the path is not to a directory.
    if not os.path.isdir(path):
        yield path

    # Walk the path if the path is to a directory.
    for root_dir, dirnames, filenames in os.walk(path):

        # Exclude git files and directories if directed.
        if not include_git:
            dirnames[:] = [d for d in dirnames if d != ".git"]

        for filename in filenames:
            yield os.path.join(root_dir, filename)


@contextmanager
def hide_cursor():
    """Hide the cursor and then finally show it again.

    Used as a decorator to hide the cursor when a function starts and
    show the cursor again when the function finishes or an exception occurs.

    Yields
    ------
    None

    """
    # Hide the cursor.
    print("\033[?25l", end="")
    try:
        yield
    finally:
        # Show the cursor.
        print("\033[?25h", end="")


def print_header(header):
    """Print a underlined header.

    Parameters
    ----------
    header : str
        The header text.

    Returns
    -------
    None

    """
    print(header)
    print("-" * len(header))


def print_progress_bar(start_time, count, total, print_frequency=100):
    """Print a progress bar.

    Parameters
    ----------
    start_time : datetime
        When the operation being reported on began.
    count : int
        The number of cycles completed.
    total : int
        The total number of cycles to complete.
    print_frequency : int, optional
        The number of cycles to complete before re-printing the count. The
        default is 100.

    Returns
    -------
    None

    """
    # Print the progress bar at the specified frequency or when the operation
    # is finished.
    is_finished = count == total
    update_now = count % print_frequency == 0

    if is_finished or update_now:

        # Build the progress bar.
        progress_bar = _build_progress_bar(count, total)
        progress_bar = _add_percentage_to_progress_bar(
            progress_bar,
            count,
            total,
        )
        progress_bar = _add_time_to_progress_bar(
            progress_bar,
            start_time,
            count,
            total,
        )

        # Print the progress bar.
        clear_line()
        print(progress_bar, end="\r")


def _add_percentage_to_progress_bar(progress_bar, count, total):
    """Add the percentage to the progress bar.

    Parameters
    ----------
    progress_bar : str
        The bar component of the progress bar, as built so far.
    count : int
        The number of cycles completed.
    total : int
        The total number of cycles to complete.

    Returns
    -------
    str
        The progress bar, with percentage appended if appropriate.

    """
    operation_is_not_finished = count != total

    if operation_is_not_finished:
        progress = count / total if total > 0 else 1
        percentage = progress * 100
        # Round the percentage down so that it is never shown as 100% before
        # the operation is finished.
        rounded_percentage = decimal.Decimal(percentage).quantize(
            decimal.Decimal("0.0"), rounding=decimal.ROUND_DOWN,
        )
        progress_bar += "  {}%".format(rounded_percentage)

    return progress_bar


def _add_time_to_progress_bar(progress_bar, start_time, count, total):
    """Add the time remaining to the progress bar.

    Parameters
    ----------
    progress_bar : str
        The bar component of the progress bar, as built so far.
    start_time : datetime
        When the operation being reported on began.
    count : int
        The number of cycles completed.
    total : int
        The total number of cycles to complete.

    Returns
    -------
    str
        The progress bar, with time remaining appended if appropriate.

    """
    operation_is_not_finished = count != total
    time_elapsed = datetime.datetime.now() - start_time
    seconds_elapsed = time_elapsed.total_seconds()
    show_time_remaining = seconds_elapsed > 2 and operation_is_not_finished

    if show_time_remaining:
        rate = count / seconds_elapsed
        seconds_remaining = (total - count) / rate
        time_remaining = _readable_duration(seconds_remaining)
        progress_bar += "  |  {} to go".format(time_remaining)

    return progress_bar


def _build_progress_bar(count, total):
    """Build the bar component of the progress bar.

    Parameters
    ----------
    count : int
        The number of cycles completed.
    total : int
        The total number of cycles to complete.

    Returns
    -------
    str
        The bar component of the progress bar.

    """
    # If total is 0, the operation has nothing to do, so it is complete.
    progress = count / total if total != 0 else 1
    bar_width = constants.TABLE_WIDTH
    progress_width = int(progress * bar_width)

    return ("█" * progress_width).ljust(bar_width, "░")


def _get_immutable_flag(path):
    """Get the immutable flag of a file.

    This function assumes that the path has already been confirmed to reference
    a file.

    Parameters
    ----------
    path : str
        An absolute path to a file.

    Returns
    -------
    str
        The string "i" if the file is immutable, or "-" if it is not.

    Raises
    ------
    GetAttributeError
        If the exit status of the lsattr command is non-zero.

    """
    try:
        lsattr_result = subprocess.run(
            ["lsattr", path],
            check=True,
            stderr=subprocess.DEVNULL,
            stdout=subprocess.PIPE,
            universal_newlines=True,
        )
    except subprocess.CalledProcessError:
        msg = "Immutable attribute could not be accessed for {}".format(path)
        raise exceptions.GetAttributeError(msg)

    # Extract the immutable attribute from the command output.
    attributes = lsattr_result.stdout.split()[0]
    immutable_flag = list(attributes)[4]

    return immutable_flag


def _readable_duration(duration):
    """Convert a duration in seconds to a human-readable duration.

    Parameters
    ----------
    duration : int
        A duration in seconds.

    Returns
    -------
    str
        A human-readable duration.

    """
    rounded_duration = round(duration)

    if rounded_duration < 60:
        duration_string = "{}s".format(rounded_duration)
    elif rounded_duration < 60 * 60:
        minutes, seconds = divmod(rounded_duration, 60)
        duration_string = "{}m {:02d}s".format(minutes, seconds)
    else:
        total_minutes, seconds = divmod(rounded_duration, 60)
        hours, minutes = divmod(total_minutes, 60)
        duration_string = (
            "{:,}h {:02d}m {:02d}s".format(hours, minutes, seconds)
        )

    return duration_string



import datetime
import os
import subprocess

from entomb import (
    exceptions,
    utilities,
)


@utilities.hide_cursor()
def process_objects(path, immutable, include_git, dry_run):
    """Set or unset the immutable attribute for all files on a path.

    Parameters
    ----------
    path : str
        An absolute path.
    immutable: bool
        Set immutable attributes if True, unset immutable attributes if False.
    include_git: bool
        Whether to include git files and directories.
    dry_run: bool
        Whether to do a dry run which makes no changes.

    Returns
    -------
    None

    Raises
    ------
    AssertionError
        If the path does not exist.

    """
    # Parameter check.
    assert os.path.exists(path)

    # Set up.
    attribute_changed_count = 0
    attribute_settable_count = 0
    errors = []
    file_count = 0
    link_count = 0
    operation = "entombed" if immutable else "unset"

    # Print the operation.
    if immutable:
        print("Entomb objects")
    else:
        print("Unset objects")
    print()

    # Print the progress header and set up the progress bar.
    utilities.print_header("Progress")
    total_file_paths = utilities.count_file_paths(path, include_git)
    start_time = datetime.datetime.now()
    utilities.print_progress_bar(start_time, 0, total_file_paths)

    # Walk the tree.
    for file_path in utilities.file_paths(path, include_git):

        # Count links, but don't try to operate on them as they don't have
        # an immutable attribute.
        if os.path.islink(file_path):
            link_count += 1

        else:
            # Change the file's attribute if necessary.
            try:
                attribute_was_changed = _process_object(
                    file_path,
                    immutable,
                    dry_run,
                )
                attribute_settable_count += 1
                if attribute_was_changed:
                    attribute_changed_count += 1
            except exceptions.SetAttributeError as error:
                errors.append(error)

            # Count the file.
            file_count += 1

        # Update the progress bar.
        utilities.print_progress_bar(
            start_time,
            (file_count + link_count),
            total_file_paths,
        )

    print()
    print()

    # Print the changes.
    if file_count > 0:
        utilities.print_header("Changes")
        print("{} {} files".format(operation.title(), attribute_changed_count))
        print()

    # Print a summary.
    utilities.print_header("Summary")
    if file_count > 0:
        print(
            "All {} files for which immutability can be set are now {}"
            .format(attribute_settable_count, operation),
        )
        print("All {} links were ignored".format(link_count))
    else:
        print("No files were found")
    print()

    # Print any errors.
    _print_errors(errors)


def _print_errors(errors):
    """Print the list of errors resulting from file processing.

    Parameters
    ----------
    errors : list of str
        A list of error messages.

    Returns
    -------
    None

    """
    # Return if there are no errors.
    if not errors:
        return

    # Print the header.
    utilities.print_header("Errors")

    # Print up to 10 errors.
    for error in errors[:10]:
        print(">> {}".format(error))

    # If there are more than 10 errors, print a message about how many more
    # there are.
    error_count = len(errors)
    if error_count > 10:
        unshown_errors = len(errors) - 10
        print(">> Plus {} more errors".format(unshown_errors))

    print()


def _process_object(path, immutable, dry_run):
    """Set or unset the immutable attribute for a file.

    Parameters
    ----------
    path : str
        The absolute path of a file.
    immutable: bool
        Set immutable attribute if True, unset immutable attribute if False.
    dry_run : bool
        Whether to do a dry run which makes no changes.

    Returns
    -------
    bool
        Whether the immutable attribute was changed, or if this was a dry run,
        should have been changed.

    Raises
    ------
    AssertionError
        If the path is a directory, is a link or does not exist.
    SetAttributeError
        If the path's immutable attribute cannot be set.

    """
    # Parameter check.
    assert not os.path.isdir(path)
    assert not os.path.islink(path)
    assert os.path.exists(path)

    try:
        is_immutable = utilities.file_is_immutable(path)
    except exceptions.GetAttributeError:
        msg = "Immutable attribute not settable for {}".format(path)
        raise exceptions.SetAttributeError(msg)

    change_attribute = immutable != is_immutable

    if change_attribute and not dry_run:
        attribute = "+i" if immutable else "-i"
        _set_attribute(attribute, path)

    # The value of change_attribute is a proxy for whether the immutable
    # attribute was changed, or if this was a dry run, should have been
    # changed.
    return change_attribute


def _set_attribute(attribute, path):
    """Set or unset an attribute for a file.

    Parameters
    ----------
    attribute: str
        The attribute to be set. In the form of "+i" or "-i".
    path : str
        The absolute path of a file.

    Returns
    -------
    None

    Raises
    ------
    SetAttributeError
        If the exit status of the chattr command is non-zero.

    """
    try:
        subprocess.run(
            ["sudo", "chattr", attribute, path],
            check=True,
            stderr=subprocess.STDOUT,
            stdout=subprocess.DEVNULL,
        )
    except subprocess.CalledProcessError:
        msg = "Immutable attribute not settable for {}".format(path)
        raise exceptions.SetAttributeError(msg)



import unittest
import unittest.mock as mock

import entomb.listing as listing
from tests import (
    constants,
    helpers,
)


class TestListing(unittest.TestCase):
    """Tests for the listing module.

    """

    def setUp(self):
        """Create temporary directories and files.

        """
        helpers.set_up()

    def test_list_files(self):
        """Test the list_files function.

        """
        # Test immutable files excluding git.
        with mock.patch("builtins.print") as mocked_print:
            listing.list_files(
                constants.DIRECTORY_PATH,
                immutable=True,
                include_git=False,
            )
        expected = [
            mock.call("\033[?25l", end=""),
            mock.call("List immutable files"),
            mock.call(),
            mock.call("Immutable files"),
            mock.call("---------------"),
            mock.call("Counting file paths: 0", end="\r"),
            mock.call("\033[K", end=""),
            mock.call("\033[K", end=""),
            mock.call(
                "░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░  0.0%",
                end="\r",
            ),
            mock.call("\033[K", end=""),
            mock.call("/tmp/entomb_testing/immutable.txt"),
            mock.call("\033[K", end=""),
            mock.call(
                "█████░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░  12.5%",
                end="\r",
            ),
            mock.call("\033[K", end=""),
            mock.call(
                "██████████░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░  25.0%",
                end="\r",
            ),
            mock.call("\033[K", end=""),
            mock.call(
                "███████████████░░░░░░░░░░░░░░░░░░░░░░░░░  37.5%",
                end="\r",
            ),
            mock.call("\033[K", end=""),
            mock.call(
                "████████████████████░░░░░░░░░░░░░░░░░░░░  50.0%",
                end="\r",
            ),
            mock.call("\033[K", end=""),
            mock.call(
                "█████████████████████████░░░░░░░░░░░░░░░  62.5%",
                end="\r",
            ),
            mock.call("\033[K", end=""),
            mock.call("/tmp/entomb_testing/subdirectory/immutable.txt"),
            mock.call("\033[K", end=""),
            mock.call(
                "██████████████████████████████░░░░░░░░░░  75.0%",
                end="\r",
            ),
            mock.call("\033[K", end=""),
            mock.call(
                "███████████████████████████████████░░░░░  87.5%",
                end="\r",
            ),
            mock.call("\033[K", end=""),
            mock.call(
                "████████████████████████████████████████",
                end="\r",
            ),
            mock.call("\033[K", end=""),
            mock.call(),
            mock.call("Summary"),
            mock.call("-------"),
            mock.call("6 files were examined"),
            mock.call("2 files are immutable"),
            mock.call(),
            mock.call("\033[?25h", end=""),
        ]
        self.assertEqual(mocked_print.mock_calls, expected)

        # Test immutable files including git.
        with mock.patch("builtins.print") as mocked_print:
            listing.list_files(
                constants.DIRECTORY_PATH,
                immutable=True,
                include_git=True,
            )
        expected = [
            mock.call("\033[?25l", end=""),
            mock.call("List immutable files"),
            mock.call(),
            mock.call("Immutable files"),
            mock.call("---------------"),
            mock.call("Counting file paths: 0", end="\r"),
            mock.call("\033[K", end=""),
            mock.call("\033[K", end=""),
            mock.call(
                "░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░  0.0%",
                end="\r",
            ),
            mock.call("\033[K", end=""),
            mock.call("/tmp/entomb_testing/immutable.txt"),
            mock.call("\033[K", end=""),
            mock.call(
                "████░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░  10.0%",
                end="\r",
            ),
            mock.call("\033[K", end=""),
            mock.call(
                "████████░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░  20.0%",
                end="\r",
            ),
            mock.call("\033[K", end=""),
            mock.call(
                "████████████░░░░░░░░░░░░░░░░░░░░░░░░░░░░  30.0%",
                end="\r",
            ),
            mock.call("\033[K", end=""),
            mock.call(
                "████████████████░░░░░░░░░░░░░░░░░░░░░░░░  40.0%",
                end="\r",
            ),
            mock.call("\033[K", end=""),
            mock.call(
                "████████████████████░░░░░░░░░░░░░░░░░░░░  50.0%",
                end="\r",
            ),
            mock.call("\033[K", end=""),
            mock.call(
                "████████████████████████░░░░░░░░░░░░░░░░  60.0%",
                end="\r",
            ),
            mock.call("\033[K", end=""),
            mock.call(
                "████████████████████████████░░░░░░░░░░░░  70.0%",
                end="\r",
            ),
            mock.call("\033[K", end=""),
            mock.call("/tmp/entomb_testing/subdirectory/immutable.txt"),
            mock.call("\033[K", end=""),
            mock.call(
                "████████████████████████████████░░░░░░░░  80.0%",
                end="\r",
            ),
            mock.call("\033[K", end=""),
            mock.call(
                "████████████████████████████████████░░░░  90.0%",
                end="\r",
            ),
            mock.call("\033[K", end=""),
            mock.call(
                "████████████████████████████████████████",
                end="\r",
            ),
            mock.call("\033[K", end=""),
            mock.call(),
            mock.call("Summary"),
            mock.call("-------"),
            mock.call("8 files were examined"),
            mock.call("2 files are immutable"),
            mock.call(),
            mock.call("\033[?25h", end=""),
        ]
        self.assertEqual(mocked_print.mock_calls, expected)

        # Test mutable files excluding git.
        with mock.patch("builtins.print") as mocked_print:
            listing.list_files(
                constants.DIRECTORY_PATH,
                immutable=False,
                include_git=False,
            )
        expected = [
            mock.call("\033[?25l", end=""),
            mock.call("List mutable files"),
            mock.call(),
            mock.call("Mutable files"),
            mock.call("-------------"),
            mock.call("Counting file paths: 0", end="\r"),
            mock.call("\033[K", end=""),
            mock.call("\033[K", end=""),
            mock.call(
                "░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░  0.0%",
                end="\r",
            ),
            mock.call("\033[K", end=""),
            mock.call(
                "█████░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░  12.5%",
                end="\r",
            ),
            mock.call("\033[K", end=""),
            mock.call(
                "██████████░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░  25.0%",
                end="\r",
            ),
            mock.call("\033[K", end=""),
            mock.call("/tmp/entomb_testing/mutable.txt"),
            mock.call("\033[K", end=""),
            mock.call(
                "███████████████░░░░░░░░░░░░░░░░░░░░░░░░░  37.5%",
                end="\r",
            ),
            mock.call("\033[K", end=""),
            mock.call(
                "████████████████████░░░░░░░░░░░░░░░░░░░░  50.0%",
                end="\r",
            ),
            mock.call("\033[K", end=""),
            mock.call(
                "█████████████████████████░░░░░░░░░░░░░░░  62.5%",
                end="\r",
            ),
            mock.call("\033[K", end=""),
            mock.call(
                "██████████████████████████████░░░░░░░░░░  75.0%",
                end="\r",
            ),
            mock.call("\033[K", end=""),
            mock.call(
                "███████████████████████████████████░░░░░  87.5%",
                end="\r",
            ),
            mock.call("\033[K", end=""),
            mock.call("/tmp/entomb_testing/subdirectory/mutable.txt"),
            mock.call("\033[K", end=""),
            mock.call(
                "████████████████████████████████████████",
                end="\r",
            ),
            mock.call("\033[K", end=""),
            mock.call(),
            mock.call("Summary"),
            mock.call("-------"),
            mock.call("6 files were examined"),
            mock.call("2 files are mutable"),
            mock.call(),
            mock.call("\033[?25h", end=""),
        ]
        self.assertEqual(mocked_print.mock_calls, expected)

        # Test mutable files including git.
        with mock.patch("builtins.print") as mocked_print:
            listing.list_files(
                constants.DIRECTORY_PATH,
                immutable=False,
                include_git=True,
            )
        expected = [
            mock.call("\033[?25l", end=""),
            mock.call("List mutable files"),
            mock.call(),
            mock.call("Mutable files"),
            mock.call("-------------"),
            mock.call("Counting file paths: 0", end="\r"),
            mock.call("\033[K", end=""),
            mock.call("\033[K", end=""),
            mock.call(
                "░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░  0.0%",
                end="\r",
            ),
            mock.call("\033[K", end=""),
            mock.call(
                "████░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░  10.0%",
                end="\r",
            ),
            mock.call("\033[K", end=""),
            mock.call(
                "████████░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░  20.0%",
                end="\r",
            ),
            mock.call("\033[K", end=""),
            mock.call("/tmp/entomb_testing/mutable.txt"),
            mock.call("\033[K", end=""),
            mock.call(
                "████████████░░░░░░░░░░░░░░░░░░░░░░░░░░░░  30.0%",
                end="\r",
            ),
            mock.call("\033[K", end=""),
            mock.call(
                "████████████████░░░░░░░░░░░░░░░░░░░░░░░░  40.0%",
                end="\r",
            ),
            mock.call("\033[K", end=""),
            mock.call(
                "████████████████████░░░░░░░░░░░░░░░░░░░░  50.0%",
                end="\r",
            ),
            mock.call("\033[K", end=""),
            mock.call("/tmp/entomb_testing/.git/mutable.txt"),
            mock.call("\033[K", end=""),
            mock.call(
                "████████████████████████░░░░░░░░░░░░░░░░  60.0%",
                end="\r",
            ),
            mock.call("\033[K", end=""),
            mock.call("/tmp/entomb_testing/.git/subdirectory/mutable.txt"),
            mock.call("\033[K", end=""),
            mock.call(
                "████████████████████████████░░░░░░░░░░░░  70.0%",
                end="\r",
            ),
            mock.call("\033[K", end=""),
            mock.call(
                "████████████████████████████████░░░░░░░░  80.0%",
                end="\r",
            ),
            mock.call("\033[K", end=""),
            mock.call(
                "████████████████████████████████████░░░░  90.0%",
                end="\r",
            ),
            mock.call("\033[K", end=""),
            mock.call("/tmp/entomb_testing/subdirectory/mutable.txt"),
            mock.call("\033[K", end=""),
            mock.call(
                "████████████████████████████████████████",
                end="\r",
            ),
            mock.call("\033[K", end=""),
            mock.call(),
            mock.call("Summary"),
            mock.call("-------"),
            mock.call("8 files were examined"),
            mock.call("4 files are mutable"),
            mock.call(),
            mock.call("\033[?25h", end=""),
        ]
        self.assertEqual(mocked_print.mock_calls, expected)

        # Test mutable files excluding git after making all files immutable.
        helpers.set_file_immutable_attribute(
            constants.GIT_SUBDIRECTORY_MUTABLE_FILE_PATH,
            immutable=True,
        )
        helpers.set_file_immutable_attribute(
            constants.MUTABLE_FILE_PATH,
            immutable=True,
        )
        helpers.set_file_immutable_attribute(
            constants.SUBDIRECTORY_MUTABLE_FILE_PATH,
            immutable=True,
        )
        with mock.patch("builtins.print") as mocked_print:
            listing.list_files(
                constants.DIRECTORY_PATH,
                immutable=False,
                include_git=False,
            )
        expected = [
            mock.call("\033[?25l", end=""),
            mock.call("List mutable files"),
            mock.call(),
            mock.call("Mutable files"),
            mock.call("-------------"),
            mock.call("Counting file paths: 0", end="\r"),
            mock.call("\033[K", end=""),
            mock.call("\033[K", end=""),
            mock.call(
                "░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░  0.0%",
                end="\r",
            ),
            mock.call("\033[K", end=""),
            mock.call(
                "█████░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░  12.5%",
                end="\r",
            ),
            mock.call("\033[K", end=""),
            mock.call(
                "██████████░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░  25.0%",
                end="\r",
            ),
            mock.call("\033[K", end=""),
            mock.call(
                "███████████████░░░░░░░░░░░░░░░░░░░░░░░░░  37.5%",
                end="\r",
            ),
            mock.call("\033[K", end=""),
            mock.call(
                "████████████████████░░░░░░░░░░░░░░░░░░░░  50.0%",
                end="\r",
            ),
            mock.call("\033[K", end=""),
            mock.call(
                "█████████████████████████░░░░░░░░░░░░░░░  62.5%",
                end="\r",
            ),
            mock.call("\033[K", end=""),
            mock.call(
                "██████████████████████████████░░░░░░░░░░  75.0%",
                end="\r",
            ),
            mock.call("\033[K", end=""),
            mock.call(
                "███████████████████████████████████░░░░░  87.5%",
                end="\r",
            ),
            mock.call("\033[K", end=""),
            mock.call(
                "████████████████████████████████████████",
                end="\r",
            ),
            mock.call("\033[K", end=""),
            mock.call("-"),
            mock.call(),
            mock.call("Summary"),
            mock.call("-------"),
            mock.call("6 files were examined"),
            mock.call("0 files are mutable"),
            mock.call(),
            mock.call("\033[?25h", end=""),
        ]
        self.assertEqual(mocked_print.mock_calls, expected)

        # Test for a non-existent path.
        with self.assertRaises(AssertionError):
            listing.list_files(
                constants.NON_EXISTENT_PATH,
                immutable=False,
                include_git=False,
            )

        # Test a named pipe.
        with mock.patch("builtins.print") as mocked_print:
            listing.list_files(
                constants.NAMED_PIPE_PATH,
                immutable=False,
                include_git=True,
            )
        expected = [
            mock.call("\033[?25l", end=""),
            mock.call("List mutable files"),
            mock.call(),
            mock.call("Mutable files"),
            mock.call("-------------"),
            mock.call("Counting file paths: 0", end="\r"),
            mock.call("\033[K", end=""),
            mock.call("\033[K", end=""),
            mock.call(
                "░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░  0.0%",
                end="\r",
            ),
            mock.call("\033[K", end=""),
            mock.call(
                "████████████████████████████████████████",
                end="\r",
            ),
            mock.call("\033[K", end=""),
            mock.call("-"),
            mock.call(),
            mock.call("Summary"),
            mock.call("-------"),
            mock.call("1 files were examined"),
            mock.call("0 files are mutable"),
            mock.call(),
            mock.call("\033[?25h", end=""),
        ]
        self.assertEqual(mocked_print.mock_calls, expected)

        # Test a file which is readable only by root.
        with mock.patch("builtins.print") as mocked_print:
            listing.list_files(
                constants.READABLE_BY_ROOT_FILE_PATH,
                immutable=True,
                include_git=False,
            )
        expected = [
            mock.call("\033[?25l", end=""),
            mock.call("List immutable files"),
            mock.call(),
            mock.call("Immutable files"),
            mock.call("---------------"),
            mock.call("Counting file paths: 0", end="\r"),
            mock.call("\033[K", end=""),
            mock.call("\033[K", end=""),
            mock.call(
                "░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░  0.0%",
                end="\r",
            ),
            mock.call("\033[K", end=""),
            mock.call(
                "████████████████████████████████████████",
                end="\r",
            ),
            mock.call("\033[K", end=""),
            mock.call("-"),
            mock.call(),
            mock.call("Summary"),
            mock.call("-------"),
            mock.call("1 files were examined"),
            mock.call("0 files are immutable"),
            mock.call(),
            mock.call("\033[?25h", end=""),
        ]
        self.assertEqual(mocked_print.mock_calls, expected)

    def test__print_the_path(self):
        """Test the _print_the_path function.

        """
        self.assertTrue(
            listing._print_the_path(
                constants.IMMUTABLE_FILE_PATH,
                immutable=True,
            ),
        )
        self.assertFalse(
            listing._print_the_path(
                constants.IMMUTABLE_FILE_PATH,
                immutable=False,
            ),
        )
        self.assertFalse(
            listing._print_the_path(
                constants.MUTABLE_FILE_PATH,
                immutable=True,
            ),
        )
        self.assertTrue(
            listing._print_the_path(
                constants.MUTABLE_FILE_PATH,
                immutable=False,
            ),
        )
        with self.assertRaises(AssertionError):
            listing._print_the_path(
                constants.NON_EXISTENT_PATH,
                immutable=True,
            )
        with self.assertRaises(AssertionError):
            listing._print_the_path(constants.DIRECTORY_PATH, immutable=False)
        with self.assertRaises(AssertionError):
            listing._print_the_path(constants.LINK_PATH, immutable=True)

    def tearDown(self):
        """Delete temporary directories and files.

        """
        helpers.tear_down()



# Package metadata
__author__ = "Sky Christensen"
__author_email__ = "sky@skychristensen.com"
__description__ = "File immutability manager"
__licence__ = "MIT"
__title__ = "Entomb"
__url__ = "https://github.com/countermeasure/entomb"
__version__ = "0.3.1"



import datetime
import os

from entomb import (
    exceptions,
    utilities,
)


@utilities.hide_cursor()
def list_files(path, immutable, include_git):
    """Print a list of the immutable or mutable files on the path.

    Parameters
    ----------
    path : str
        An absolute path.
    immutable: bool
        List immutable files if True, mutable files if False.
    include_git: bool
        Whether to include git files and directories.

    Returns
    -------
    None

    Raises
    ------
    AssertionError
        If the path does not exist.

    """
    # Parameter check.
    assert os.path.exists(path)

    # Set up.
    file_count = 0
    link_count = 0
    printed_file_count = 0
    state = "immutable" if immutable else "mutable"

    # Print the operation.
    print("List {} files".format(state))
    print()

    # Print the list header.
    list_header = "{} files".format(state.title())
    utilities.print_header(list_header)

    # Set up the progress bar.
    total_file_paths = utilities.count_file_paths(path, include_git)
    start_time = datetime.datetime.now()
    utilities.print_progress_bar(start_time, 0, total_file_paths)

    # Walk the tree.
    for file_path in utilities.file_paths(path, include_git):

        # Count the link.
        if os.path.islink(file_path):
            link_count += 1

        # Count the file and print its path if appropriate.
        else:
            if _print_the_path(file_path, immutable):
                printed_file_count += 1
                utilities.clear_line()
                print(file_path)

            file_count += 1

        # Update the progress bar.
        utilities.print_progress_bar(
            start_time,
            (file_count + link_count),
            total_file_paths,
            1,
        )

    # Clear the final progress message.
    utilities.clear_line()

    # If the list is empty, print an empty indicator.
    if printed_file_count == 0:
        print("-")

    # Print a summary.
    print()
    utilities.print_header("Summary")
    print("{} files were examined".format(file_count))
    print("{} files are {}".format(printed_file_count, state))
    print()


def _print_the_path(path, immutable):
    """Determine whether to print the file path.

    The path should be printed if the immutable parameter is True and the file
    is immutable, or if the immutable parameter is False and the file is
    mutable.

    The path should not be printed if the immutability attribute can't be
    accessed.

    Parameters
    ----------
    path : str
        An absolute path.
    immutable: bool
        List immutable files if True, mutable files if False.

    Returns
    -------
    bool
        Whether to print the file path.

    Raises
    ------
    AssertionError
        If the path is a directory, is a link or does not exist.

    """
    # Parameter check.
    assert not os.path.isdir(path)
    assert not os.path.islink(path)
    assert os.path.exists(path)
    try:
        is_immutable = utilities.file_is_immutable(path)
    except exceptions.GetAttributeError:
        return False

    print_immutable_file_path = is_immutable and immutable
    print_mutable_file_path = not is_immutable and not immutable

    return print_immutable_file_path or print_mutable_file_path



import datetime
import math
import os

from entomb import (
    constants,
    exceptions,
    utilities,
)


@utilities.hide_cursor()
def produce_report(path, include_git):
    """Print a report.

    Parameters
    ----------
    path : str
        An absolute path.
    include_git: bool
        Whether to include git files and directories.

    Returns
    -------
    None

    Raises
    ------
    AssertionError
        If the path does not exist.

    """
    # Parameter check.
    assert os.path.exists(path)

    # Set up.
    directory_count = 0
    immutable_file_count = 0
    inaccessible_file_count = 0
    link_count = 0
    mutable_file_count = 0

    # Print the operation.
    print("Produce report")
    print()

    # If the path is not a directory, print an abbreviated report then return.
    if not os.path.isdir(path):
        _print_abbreviated_report(path)
        return

    # Print the progress header and set up the progress bar.
    utilities.print_header("Progress")
    total_file_paths = utilities.count_file_paths(path, include_git)
    start_time = datetime.datetime.now()
    utilities.print_progress_bar(start_time, 0, total_file_paths)

    # Walk the tree.
    for root_dir, dirnames, filenames in os.walk(path):

        # Exclude git files and directories if directed.
        if not include_git:
            dirnames[:] = [d for d in dirnames if d != ".git"]

        # Count the directory.
        directory_count += 1

        # Examine each file path.
        for filename in filenames:
            file_path = os.path.join(root_dir, filename)

            # Count the link.
            if os.path.islink(file_path):
                link_count += 1

            # Count the file.
            else:
                try:
                    if utilities.file_is_immutable(file_path):
                        immutable_file_count += 1
                    else:
                        mutable_file_count += 1
                except exceptions.GetAttributeError:
                    inaccessible_file_count += 1

            # Update the progress bar.
            total_count = (
                immutable_file_count
                + inaccessible_file_count
                + link_count
                + mutable_file_count
            )
            utilities.print_progress_bar(
                start_time,
                total_count,
                total_file_paths,
            )

    print()
    print()

    _print_full_report(
        directory_count,
        link_count,
        immutable_file_count,
        inaccessible_file_count,
        mutable_file_count,
    )


def _print_abbreviated_report(path):
    """Print a report for a path which is not a directory.

    Parameters
    ----------
    path : str
        An absolute path.

    Returns
    -------
    None

    Raises
    ------
    AssertionError
        If the path is a directory or does not exist.

    """
    # Parameter check.
    assert not os.path.isdir(path)
    assert os.path.exists(path)

    utilities.print_header("Report")

    if os.path.islink(path):
        print("A link has no immutable attribute")
    else:
        try:
            if utilities.file_is_immutable(path):
                print("File is immutable")
            else:
                print("File is mutable")
        except exceptions.GetAttributeError:
            print("Immutable attribute could not be accessed")

    print()


def _print_full_report(directory_count, link_count, immutable_file_count,
                       inaccessible_file_count, mutable_file_count):
    """Print a report for a path which is a file or a link.

    Parameters
    ----------
    directory_count : int
        The number of directories counted.
    link_count : int
        The number of links counted.
    immutable_file_count : int
        The number of immutable files counted.
    inaccessible_file_count : int
        The number of files for which the immutability attribute could not be
        accessed.
    mutable_file_count : int
        The number of mutable files counted.

    Returns
    -------
    None

    """
    # Do calculations.
    subdirectory_count = directory_count - 1
    total_file_count = (
        immutable_file_count
        + inaccessible_file_count
        + mutable_file_count
    )

    try:
        entombed_proportion = immutable_file_count / total_file_count
        entombed_percentage_integer = math.floor(entombed_proportion * 100)
        entombed_percentage = "{}%".format(entombed_percentage_integer)
    except ZeroDivisionError:
        entombed_percentage = "n/a"

    # Print the report.
    report_separator = "-" * constants.TABLE_WIDTH
    print("Report")
    print(report_separator)
    print(
        "Immutable files",
        "{:,}".format(immutable_file_count).rjust(constants.TABLE_WIDTH - 16),
    )
    print(report_separator)
    print(
        "Mutable files",
        "{:,}".format(mutable_file_count).rjust(constants.TABLE_WIDTH - 14),
    )
    print(report_separator)
    if inaccessible_file_count:
        print(
            "Inaccessible files",
            "{:,}".format(inaccessible_file_count).
            rjust(constants.TABLE_WIDTH - 19),
        )
        print(report_separator)
    print(
        "All files",
        "{:,}".format(total_file_count).rjust(constants.TABLE_WIDTH - 10),
    )
    print(report_separator)
    print(
        "Entombed",
        entombed_percentage.rjust(constants.TABLE_WIDTH - 9),
    )
    print(report_separator)
    print(
        "Links",
        "{:,}".format(link_count).rjust(constants.TABLE_WIDTH - 6),
    )
    print(report_separator)
    print(
        "Sub-directories",
        "{:,}".format(subdirectory_count).rjust(constants.TABLE_WIDTH - 16),
    )
    print(report_separator)
    print()



import argparse
import io
import os
import signal
import subprocess
import unittest
import unittest.mock as mock

import entomb.core as core
from tests import (
    constants,
    helpers,
)


EXECUTABLE = "entomb"


class TestCore(unittest.TestCase):
    """Tests for the core module.

    """

    def setUp(self):
        """Create temporary directories and files.

        """
        helpers.set_up()

    def test_main(self):
        """Test the main function.

        """
        # Test invoking with conflicting arguments.
        with mock.patch("builtins.print") as mocked_print:
            core.main([
                EXECUTABLE,
                constants.DIRECTORY_PATH,
                "-d",
                "--list-immutable",
            ])
        expected = [
            mock.call(),
            mock.call(
                "ERROR:",
                "--dry-run and --list-immutable cannot be passed together",
            ),
        ]
        self.assertEqual(mocked_print.mock_calls, expected)

        # Test invoking with a non-existent path as the target.
        with mock.patch("builtins.print") as mocked_print:
            core.main([EXECUTABLE, constants.NON_EXISTENT_PATH])
        expected = [
            mock.call(),
            mock.call("ERROR: The path you gave could not be found"),
            mock.call(
                ">> You entered this path:",
                "/a/path/that/does/not/exist",
            ),
        ]
        self.assertEqual(mocked_print.mock_calls, expected)

        # Test generating a report for a directory including git files.
        with mock.patch("builtins.print") as mocked_print:
            core.main([EXECUTABLE, constants.DIRECTORY_PATH, "-r"])
        expected = [
            mock.call(),
            mock.call("======"),
            mock.call("Entomb"),
            mock.call("======"),
            mock.call(),
            mock.call("Directory"),
            mock.call("---------"),
            mock.call("/tmp/entomb_testing"),
            mock.call(),
            mock.call("Operation"),
            mock.call("---------"),
            mock.call("\033[?25l", end=""),
            mock.call("Produce report"),
            mock.call(),
            mock.call("Progress"),
            mock.call("--------"),
            mock.call("Counting file paths: 0", end="\r"),
            mock.call("\033[K", end=""),
            mock.call("\033[K", end=""),
            mock.call(
                "░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░  0.0%",
                end="\r",
            ),
            mock.call("\033[K", end=""),
            mock.call(
                "████████████████████████████████████████",
                end="\r",
            ),
            mock.call(),
            mock.call(),
            mock.call("Report"),
            mock.call("----------------------------------------"),
            mock.call("Immutable files", "                       2"),
            mock.call("----------------------------------------"),
            mock.call("Mutable files", "                         2"),
            mock.call("----------------------------------------"),
            mock.call("Inaccessible files", "                    2"),
            mock.call("----------------------------------------"),
            mock.call("All files", "                             6"),
            mock.call("----------------------------------------"),
            mock.call("Entombed", "                            33%"),
            mock.call("----------------------------------------"),
            mock.call("Links", "                                 2"),
            mock.call("----------------------------------------"),
            mock.call("Sub-directories", "                       2"),
            mock.call("----------------------------------------"),
            mock.call(),
            mock.call("\033[?25h", end=""),
        ]
        self.assertEqual(mocked_print.mock_calls, expected)

        # Test printing a list of immutable files.
        with mock.patch("builtins.print") as mocked_print:
            core.main(
                [EXECUTABLE, constants.SUBDIRECTORY_PATH, "--list-immutable"],
            )
        expected = [
            mock.call(),
            mock.call("======"),
            mock.call("Entomb"),
            mock.call("======"),
            mock.call(),
            mock.call("Directory"),
            mock.call("---------"),
            mock.call("/tmp/entomb_testing/subdirectory"),
            mock.call(),
            mock.call("Operation"),
            mock.call("---------"),
            mock.call("\033[?25l", end=""),
            mock.call("List immutable files"),
            mock.call(),
            mock.call("Immutable files"),
            mock.call("---------------"),
            mock.call("Counting file paths: 0", end="\r"),
            mock.call("\033[K", end=""),
            mock.call("\033[K", end=""),
            mock.call(
                "░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░  0.0%",
                end="\r",
            ),
            mock.call("\033[K", end=""),
            mock.call("/tmp/entomb_testing/subdirectory/immutable.txt"),
            mock.call("\033[K", end=""),
            mock.call(
                "█████████████░░░░░░░░░░░░░░░░░░░░░░░░░░░  33.3%",
                end="\r",
            ),
            mock.call("\033[K", end=""),
            mock.call(
                "██████████████████████████░░░░░░░░░░░░░░  66.6%",
                end="\r",
            ),
            mock.call("\033[K", end=""),
            mock.call(
                "████████████████████████████████████████",
                end="\r",
            ),
            mock.call("\033[K", end=""),
            mock.call(),
            mock.call("Summary"),
            mock.call("-------"),
            mock.call("2 files were examined"),
            mock.call("1 files are immutable"),
            mock.call(),
            mock.call("\033[?25h", end=""),
        ]
        self.assertEqual(mocked_print.mock_calls, expected)

        # Test printing a list of mutable files including git files.
        with mock.patch("builtins.print") as mocked_print:
            core.main(
                [EXECUTABLE, constants.DIRECTORY_PATH, "-g", "--list-mutable"],
            )
        expected = [
            mock.call(),
            mock.call("======"),
            mock.call("Entomb"),
            mock.call("======"),
            mock.call(),
            mock.call(">> Files in .git directories are being included"),
            mock.call(),
            mock.call("Directory"),
            mock.call("---------"),
            mock.call("/tmp/entomb_testing"),
            mock.call(),
            mock.call("Operation"),
            mock.call("---------"),
            mock.call("\033[?25l", end=""),
            mock.call("List mutable files"),
            mock.call(),
            mock.call("Mutable files"),
            mock.call("-------------"),
            mock.call("Counting file paths: 0", end="\r"),
            mock.call("\033[K", end=""),
            mock.call("\033[K", end=""),
            mock.call(
                "░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░  0.0%",
                end="\r",
            ),
            mock.call("\033[K", end=""),
            mock.call(
                "████░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░  10.0%",
                end="\r",
            ),
            mock.call("\033[K", end=""),
            mock.call(
                "████████░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░  20.0%",
                end="\r",
            ),
            mock.call("\033[K", end=""),
            mock.call("/tmp/entomb_testing/mutable.txt"),
            mock.call("\033[K", end=""),
            mock.call(
                "████████████░░░░░░░░░░░░░░░░░░░░░░░░░░░░  30.0%",
                end="\r",
            ),
            mock.call("\033[K", end=""),
            mock.call(
                "████████████████░░░░░░░░░░░░░░░░░░░░░░░░  40.0%",
                end="\r",
            ),
            mock.call("\033[K", end=""),
            mock.call(
                "████████████████████░░░░░░░░░░░░░░░░░░░░  50.0%",
                end="\r",
            ),
            mock.call("\033[K", end=""),
            mock.call("/tmp/entomb_testing/.git/mutable.txt"),
            mock.call("\033[K", end=""),
            mock.call(
                "████████████████████████░░░░░░░░░░░░░░░░  60.0%",
                end="\r",
            ),
            mock.call("\033[K", end=""),
            mock.call("/tmp/entomb_testing/.git/subdirectory/mutable.txt"),
            mock.call("\033[K", end=""),
            mock.call(
                "████████████████████████████░░░░░░░░░░░░  70.0%",
                end="\r",
            ),
            mock.call("\033[K", end=""),
            mock.call(
                "████████████████████████████████░░░░░░░░  80.0%",
                end="\r",
            ),
            mock.call("\033[K", end=""),
            mock.call(
                "████████████████████████████████████░░░░  90.0%",
                end="\r",
            ),
            mock.call("\033[K", end=""),
            mock.call("/tmp/entomb_testing/subdirectory/mutable.txt"),
            mock.call("\033[K", end=""),
            mock.call(
                "████████████████████████████████████████",
                end="\r",
            ),
            mock.call("\033[K", end=""),
            mock.call(),
            mock.call("Summary"),
            mock.call("-------"),
            mock.call("8 files were examined"),
            mock.call("4 files are mutable"),
            mock.call(),
            mock.call("\033[?25h", end=""),
        ]
        self.assertEqual(mocked_print.mock_calls, expected)

        # Test a dry run for making all files in a directory immutable.
        with mock.patch("builtins.print") as mocked_print:
            core.main([EXECUTABLE, constants.DIRECTORY_PATH, "--dry-run"])
        expected = [
            mock.call(),
            mock.call("======"),
            mock.call("Entomb"),
            mock.call("======"),
            mock.call(),
            mock.call(">> This is a dry run so no changes will be made"),
            mock.call(),
            mock.call("Directory"),
            mock.call("---------"),
            mock.call("/tmp/entomb_testing"),
            mock.call(),
            mock.call("Operation"),
            mock.call("---------"),
            mock.call("\033[?25l", end=""),
            mock.call("Entomb objects"),
            mock.call(),
            mock.call("Progress"),
            mock.call("--------"),
            mock.call("Counting file paths: 0", end="\r"),
            mock.call("\033[K", end=""),
            mock.call("\033[K", end=""),
            mock.call(
                "░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░  0.0%",
                end="\r",
            ),
            mock.call("\033[K", end=""),
            mock.call(
                "████████████████████████████████████████",
                end="\r",
            ),
            mock.call(),
            mock.call(),
            mock.call("Changes"),
            mock.call("-------"),
            mock.call("Entombed 2 files"),
            mock.call(),
            mock.call("Summary"),
            mock.call("-------"),
            mock.call(
                "All 4 files for which immutability can be set are now "
                "entombed",
            ),
            mock.call("All 2 links were ignored"),
            mock.call(),
            mock.call("Errors"),
            mock.call("------"),
            mock.call(
                ">> Immutable attribute not settable for "
                "/tmp/entomb_testing/fifo",
            ),
            mock.call(
                ">> Immutable attribute not settable for "
                "/tmp/entomb_testing/readable_by_root.txt",
            ),
            mock.call(),
            mock.call("\033[?25h", end=""),
        ]
        self.assertEqual(mocked_print.mock_calls, expected)

        # Test making an immutable file mutable.
        with mock.patch("builtins.print") as mocked_print:
            core.main([EXECUTABLE, constants.IMMUTABLE_FILE_PATH, "--unset"])
        expected = [
            mock.call(),
            mock.call("======"),
            mock.call("Entomb"),
            mock.call("======"),
            mock.call(),
            mock.call("File"),
            mock.call("----"),
            mock.call("/tmp/entomb_testing/immutable.txt"),
            mock.call(),
            mock.call("Operation"),
            mock.call("---------"),
            mock.call("\033[?25l", end=""),
            mock.call("Unset objects"),
            mock.call(),
            mock.call("Progress"),
            mock.call("--------"),
            mock.call("Counting file paths: 0", end="\r"),
            mock.call("\033[K", end=""),
            mock.call("\033[K", end=""),
            mock.call(
                "░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░  0.0%",
                end="\r",
            ),
            mock.call("\033[K", end=""),
            mock.call(
                "████████████████████████████████████████",
                end="\r",
            ),
            mock.call(),
            mock.call(),
            mock.call("Changes"),
            mock.call("-------"),
            mock.call("Unset 1 files"),
            mock.call(),
            mock.call("Summary"),
            mock.call("-------"),
            mock.call(
                "All 1 files for which immutability can be set are now unset",
            ),
            mock.call("All 0 links were ignored"),
            mock.call(),
            mock.call("\033[?25h", end=""),
        ]
        self.assertEqual(mocked_print.mock_calls, expected)

    def test__check_argument_conflicts(self):
        """Test the _check_argument_conflicts function.

        """
        # Test --dry-run and --report being passed in together.
        args = argparse.Namespace(
            dry_run=True,
            include_git=False,
            list_immutable=False,
            list_mutable=False,
            path=constants.DIRECTORY_PATH,
            report=True,
            unset=False,
        )
        actual = core._check_argument_conflicts(args)
        expected = ["--dry-run and --report cannot be passed together"]
        self.assertEqual(actual, expected)

        # Test --report and --unset being passed in together.
        args = argparse.Namespace(
            dry_run=False,
            include_git=False,
            list_immutable=False,
            list_mutable=False,
            path=constants.DIRECTORY_PATH,
            report=True,
            unset=True,
        )
        actual = core._check_argument_conflicts(args)
        expected = ["--report and --unset cannot be passed together"]
        self.assertEqual(actual, expected)

        # Test --list-immutable and --unset being passed in together.
        args = argparse.Namespace(
            dry_run=False,
            include_git=False,
            list_immutable=True,
            list_mutable=False,
            path=constants.DIRECTORY_PATH,
            report=False,
            unset=True,
        )
        actual = core._check_argument_conflicts(args)
        expected = ["--list-immutable and --unset cannot be passed together"]
        self.assertEqual(actual, expected)

        # Test --list-immutable, --list-mutable and --report being passed in
        # together.
        args = argparse.Namespace(
            dry_run=False,
            include_git=False,
            list_immutable=True,
            list_mutable=True,
            path=constants.DIRECTORY_PATH,
            report=True,
            unset=False,
        )
        actual = core._check_argument_conflicts(args)
        expected = [
            "--list-immutable and --list-mutable cannot be passed together",
            "--list-immutable and --report cannot be passed together",
            "--list-mutable and --report cannot be passed together",
        ]
        self.assertEqual(actual, expected)

        # Test --dry-run, --include-git, --list-mutable and --unset being
        # passed in together.
        args = argparse.Namespace(
            dry_run=True,
            include_git=True,
            list_immutable=False,
            list_mutable=True,
            path=constants.DIRECTORY_PATH,
            report=False,
            unset=True,
        )
        actual = core._check_argument_conflicts(args)
        expected = [
            "--dry-run and --list-mutable cannot be passed together",
            "--list-mutable and --unset cannot be passed together",
        ]
        self.assertEqual(actual, expected)

        # Test --unset be passed in alone.
        args = argparse.Namespace(
            dry_run=False,
            include_git=False,
            list_immutable=False,
            list_mutable=False,
            path=constants.DIRECTORY_PATH,
            report=False,
            unset=True,
        )
        actual = core._check_argument_conflicts(args)
        expected = []
        self.assertEqual(actual, expected)

        # Test no optional arguments being passed in.
        args = argparse.Namespace(
            dry_run=False,
            include_git=False,
            list_immutable=False,
            list_mutable=False,
            path=constants.DIRECTORY_PATH,
            report=False,
            unset=False,
        )
        actual = core._check_argument_conflicts(args)
        expected = []
        self.assertEqual(actual, expected)

    def test__get_absolute_path(self):
        """Test the _get_absolute_path function.

        """
        # Test an absolute path.
        actual = core._get_absolute_path(constants.DIRECTORY_PATH)
        expected = constants.DIRECTORY_PATH
        self.assertEqual(actual, expected)

        # Test a relative path.
        actual = core._get_absolute_path(constants.RELATIVE_PATH)
        expected = os.path.join(os.getcwd(), constants.RELATIVE_PATH)
        self.assertEqual(actual, expected)

        # Test a path which contains a ".." reference.
        actual = core._get_absolute_path("/tmp/entomb_testing/..")
        expected = "/tmp"
        self.assertEqual(actual, expected)

        # Test a path which contains "." references.
        actual = core._get_absolute_path("/tmp/././entomb_testing")
        expected = "/tmp/entomb_testing"
        self.assertEqual(actual, expected)

    def test__get_path_type(self):
        """Test the _get_path_type function.

        """
        # Test a link.
        actual = core._get_path_type(constants.LINK_PATH)
        expected = "link"
        self.assertEqual(actual, expected)

        # Test a file.
        actual = core._get_path_type(constants.IMMUTABLE_FILE_PATH)
        expected = "file"
        self.assertEqual(actual, expected)

        # Test a directory.
        actual = core._get_path_type(constants.DIRECTORY_PATH)
        expected = "directory"
        self.assertEqual(actual, expected)

        # Test a named pipe.
        actual = core._get_path_type(constants.NAMED_PIPE_PATH)
        expected = "path"
        self.assertEqual(actual, expected)

        # Test a file which is readable only by root
        actual = core._get_path_type(constants.READABLE_BY_ROOT_FILE_PATH)
        expected = "file"
        self.assertEqual(actual, expected)

        # Test a relative path.
        with self.assertRaises(AssertionError):
            core._get_path_type(constants.RELATIVE_PATH)

        # Test a path which doesn't exist.
        with self.assertRaises(AssertionError):
            core._get_path_type(constants.NON_EXISTENT_PATH)

    def test__parse_args(self):
        """Test the _parse_args function.

        """
        # Test just a directory path.
        actual = core._parse_args([EXECUTABLE, constants.DIRECTORY_PATH])
        expected = argparse.Namespace(
            dry_run=False,
            include_git=False,
            list_immutable=False,
            list_mutable=False,
            path=constants.DIRECTORY_PATH,
            report=False,
            unset=False,
        )
        self.assertEqual(actual, expected)

        # Test a link path with --list--immutable and --u.
        actual = core._parse_args(
            [EXECUTABLE, constants.LINK_PATH, "--list-immutable", "-u"],
        )
        expected = argparse.Namespace(
            dry_run=False,
            include_git=False,
            list_immutable=True,
            list_mutable=False,
            path=constants.LINK_PATH,
            report=False,
            unset=True,
        )
        self.assertEqual(actual, expected)

        # Test --version without a path.
        output = io.StringIO
        with mock.patch("sys.stdout", new_callable=output) as mocked_stdout:
            # Stop stderr being displayed.
            with mock.patch("sys.stderr") as _:
                # Stop the test exiting before it can complete.
                with mock.patch("sys.exit") as _:
                    core._parse_args([EXECUTABLE, "--version"])
        actual = mocked_stdout.getvalue()
        self.assertRegex(actual, r"^Entomb \d+\.\d+\.\d+\n$")

    def test__print_argument_conflict_errors(self):
        """Test the _print_argument_conflict_errors function.

        """
        # Test a list with two error messages.
        with mock.patch("builtins.print") as mocked_print:
            core._print_argument_conflict_errors(
                ["Message one", "Message two"],
            )
        expected = [
            mock.call(),
            mock.call("ERROR:", "Message one"),
            mock.call("ERROR:", "Message two"),
        ]
        self.assertEqual(mocked_print.mock_calls, expected)

        # Test a list with no error messages.
        with mock.patch("builtins.print") as mocked_print:
            core._print_argument_conflict_errors([])
        expected = [
            mock.call(),
        ]
        self.assertEqual(mocked_print.mock_calls, expected)

    def test__print_path_not_found_error(self):
        """Test the _print_path_not_found_error function.

        """
        # Test a relative and absolute path.
        with mock.patch("builtins.print") as mocked_print:
            core._print_path_not_found_error(
                "immutable.txt",
                constants.IMMUTABLE_FILE_PATH,
            )
        expected = [
            mock.call(),
            mock.call("ERROR: The path you gave could not be found"),
            mock.call(">> You entered this path:", "immutable.txt"),
            mock.call(
                ">> It was expanded to this:",
                constants.IMMUTABLE_FILE_PATH,
            ),
        ]
        self.assertEqual(mocked_print.mock_calls, expected)

        # Test an absolute path.
        with mock.patch("builtins.print") as mocked_print:
            core._print_path_not_found_error(
                constants.IMMUTABLE_FILE_PATH,
                constants.IMMUTABLE_FILE_PATH,
            )
        expected = [
            mock.call(),
            mock.call("ERROR: The path you gave could not be found"),
            mock.call(
                ">> You entered this path:",
                constants.IMMUTABLE_FILE_PATH,
            ),
        ]
        self.assertEqual(mocked_print.mock_calls, expected)

    def test__prompt_for_sudo(self):
        """Test the _prompt_for_sudo function.

        """
        # Get as close as possible to testing for a user without root
        # privileges by using a mock side effect to print the subprocess.run
        # call.
        with mock.patch("builtins.print") as mocked_print:
            with mock.patch("subprocess.run") as mocked_run:
                mocked_run.side_effect = print
                core._prompt_for_sudo()
        expected = [
            mock.call(),
            mock.call("** To set or unset files' immutable attributes, root"),
            mock.call("** privileges are required."),
            mock.call(),
            mock.call(["sudo", "-v"], check=True, stderr=-2, stdout=-3),
            mock.call(),
        ]
        self.assertEqual(mocked_print.mock_calls, expected)

    def test__signal_handler(self):
        """Test the _signal_handler function.

        """
        # Test a keyboard interrupt signal.
        with mock.patch("builtins.print") as mocked_print:
            with mock.patch("sys.exit") as mocked_exit:
                mocked_exit.side_effect = print
                core._signal_handler(signal.SIGINT, None)
        expected = [
            mock.call(),
            mock.call(),
            mock.call(">> Keyboard interrupt signal received. Exiting."),
            mock.call(),
            mock.call(1),
        ]
        self.assertEqual(mocked_print.mock_calls, expected)

        # Test a termination signal.
        with mock.patch("builtins.print") as mocked_print:
            with mock.patch("sys.exit") as mocked_exit:
                mocked_exit.side_effect = print
                core._signal_handler(signal.SIGTERM, None)
        expected = [
            mock.call(),
            mock.call(),
            mock.call(">> Termination signal received. Exiting."),
            mock.call(),
            mock.call(1),
        ]
        self.assertEqual(mocked_print.mock_calls, expected)

    def test__user_has_root_privileges(self):
        """Test the _user_has_root_privileges function.

        """
        # Test for a user with root privileges.
        self.assertTrue(core._user_has_root_privileges())

        # Test for a user without root privileges by mocking a
        # CalledProcessError.
        with mock.patch("subprocess.run") as mocked_run:
            mocked_run.side_effect = subprocess.CalledProcessError(
                cmd=["sudo", "-ln"],
                returncode=1,
            )
            self.assertFalse(core._user_has_root_privileges())

    def tearDown(self):
        """Delete temporary directories and files.

        """
        helpers.tear_down()



import os


# Main directory paths.
DIRECTORY_PATH = "/tmp/entomb_testing"
IMMUTABLE_FILE_PATH = os.path.join(DIRECTORY_PATH, "immutable.txt")
LINK_PATH = os.path.join(DIRECTORY_PATH, "link.txt")
MUTABLE_FILE_PATH = os.path.join(DIRECTORY_PATH, "mutable.txt")
NAMED_PIPE_PATH = os.path.join(DIRECTORY_PATH, "fifo")
READABLE_BY_ROOT_FILE_PATH = os.path.join(
    DIRECTORY_PATH,
    "readable_by_root.txt",
)


# Subdirectory paths.
EMPTY_SUBDIRECTORY_PATH = os.path.join(DIRECTORY_PATH, "empty_directory")
SUBDIRECTORY_PATH = os.path.join(DIRECTORY_PATH, "subdirectory")
SUBDIRECTORY_IMMUTABLE_FILE_PATH = os.path.join(
    SUBDIRECTORY_PATH,
    "immutable.txt",
)
SUBDIRECTORY_LINK_PATH = os.path.join(SUBDIRECTORY_PATH, "link.txt")
SUBDIRECTORY_MUTABLE_FILE_PATH = os.path.join(SUBDIRECTORY_PATH, "mutable.txt")


# Fake git repo paths.
GIT_DIRECTORY_PATH = os.path.join(DIRECTORY_PATH, ".git")
GIT_DIRECTORY_MUTABLE_FILE_PATH = os.path.join(
    GIT_DIRECTORY_PATH,
    "mutable.txt",
)
GIT_SUBDIRECTORY_PATH = os.path.join(GIT_DIRECTORY_PATH, "subdirectory")
GIT_SUBDIRECTORY_MUTABLE_FILE_PATH = os.path.join(
    GIT_SUBDIRECTORY_PATH,
    "mutable.txt",
)


# Non-existent paths.
NON_EXISTENT_PATH = "/a/path/that/does/not/exist"
NON_PATH_STRING = "a non-path string"
RELATIVE_PATH = "relative/path"



import contextlib
import os
import subprocess
import unittest
import unittest.mock as mock

import entomb.exceptions as exceptions
import entomb.processing as processing
from tests import (
    constants,
    helpers,
)


class TestProcessing(unittest.TestCase):
    """Tests for the processing module.

    """

    def setUp(self):
        """Create temporary directories and files.

        """
        helpers.set_up()

    def test_process_objects(self):
        """Test the process_objects function.

        """
        # Test making files immutable excluding git files
        with mock.patch("builtins.print") as mocked_print:
            processing.process_objects(
                constants.DIRECTORY_PATH,
                immutable=True,
                include_git=False,
                dry_run=False,
            )
        expected = [
            mock.call("\033[?25l", end=""),
            mock.call("Entomb objects"),
            mock.call(),
            mock.call("Progress"),
            mock.call("--------"),
            mock.call("Counting file paths: 0", end="\r"),
            mock.call("\033[K", end=""),
            mock.call("\033[K", end=""),
            mock.call(
                "░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░  0.0%",
                end="\r",
            ),
            mock.call("\033[K", end=""),
            mock.call(
                "████████████████████████████████████████",
                end="\r",
            ),
            mock.call(),
            mock.call(),
            mock.call("Changes"),
            mock.call("-------"),
            mock.call("Entombed 2 files"),
            mock.call(),
            mock.call("Summary"),
            mock.call("-------"),
            mock.call(
                "All 4 files for which immutability can be set are now "
                "entombed",
            ),
            mock.call("All 2 links were ignored"),
            mock.call(),
            mock.call("Errors"),
            mock.call("------"),
            mock.call(
                ">> Immutable attribute not settable for "
                "/tmp/entomb_testing/fifo",
            ),
            mock.call(
                ">> Immutable attribute not settable for "
                "/tmp/entomb_testing/readable_by_root.txt",
            ),
            mock.call(),
            mock.call("\033[?25h", end=""),
        ]
        self.assertEqual(mocked_print.mock_calls, expected)

        # Confirm that all files for which the immutable attribute is settable
        # were made immutable.
        for root_dir, _, filenames in os.walk(constants.DIRECTORY_PATH):
            file_should_be_immutable = ".git" not in root_dir
            for filename in filenames:
                file_path = os.path.join(root_dir, filename)
                with contextlib.suppress(subprocess.CalledProcessError):
                    self.assertEqual(
                        helpers.file_is_immutable(file_path),
                        file_should_be_immutable,
                    )

        # Test making files mutable including git.
        with mock.patch("builtins.print") as mocked_print:
            processing.process_objects(
                constants.DIRECTORY_PATH,
                immutable=False,
                include_git=True,
                dry_run=False,
            )
        expected = [
            mock.call("\033[?25l", end=""),
            mock.call("Unset objects"),
            mock.call(),
            mock.call("Progress"),
            mock.call("--------"),
            mock.call("Counting file paths: 0", end="\r"),
            mock.call("\033[K", end=""),
            mock.call("\033[K", end=""),
            mock.call(
                "░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░  0.0%",
                end="\r",
            ),
            mock.call("\033[K", end=""),
            mock.call(
                "████████████████████████████████████████",
                end="\r",
            ),
            mock.call(),
            mock.call(),
            mock.call("Changes"),
            mock.call("-------"),
            mock.call("Unset 4 files"),
            mock.call(),
            mock.call("Summary"),
            mock.call("-------"),
            mock.call(
                "All 6 files for which immutability can be set are now unset",
            ),
            mock.call("All 2 links were ignored"),
            mock.call(),
            mock.call("Errors"),
            mock.call("------"),
            mock.call(
                ">> Immutable attribute not settable for "
                "/tmp/entomb_testing/fifo",
            ),
            mock.call(
                ">> Immutable attribute not settable for "
                "/tmp/entomb_testing/readable_by_root.txt",
            ),
            mock.call(),
            mock.call("\033[?25h", end=""),
        ]
        self.assertEqual(mocked_print.mock_calls, expected)

        # Confirm that all files for which the immutable attribute is settable
        # were made mutable.
        for root_dir, _, filenames in os.walk(constants.DIRECTORY_PATH):
            for filename in filenames:
                file_path = os.path.join(root_dir, filename)
                with contextlib.suppress(subprocess.CalledProcessError):
                    self.assertFalse(helpers.file_is_immutable(file_path))

        # Test a dry-run for making files immutable excluding git.
        with mock.patch("builtins.print") as mocked_print:
            processing.process_objects(
                constants.DIRECTORY_PATH,
                immutable=True,
                include_git=False,
                dry_run=True,
            )
        expected = [
            mock.call("\033[?25l", end=""),
            mock.call("Entomb objects"),
            mock.call(),
            mock.call("Progress"),
            mock.call("--------"),
            mock.call("Counting file paths: 0", end="\r"),
            mock.call("\033[K", end=""),
            mock.call("\033[K", end=""),
            mock.call(
                "░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░  0.0%",
                end="\r",
            ),
            mock.call("\033[K", end=""),
            mock.call(
                "████████████████████████████████████████",
                end="\r",
            ),
            mock.call(),
            mock.call(),
            mock.call("Changes"),
            mock.call("-------"),
            mock.call("Entombed 4 files"),
            mock.call(),
            mock.call("Summary"),
            mock.call("-------"),
            mock.call(
                "All 4 files for which immutability can be set are now "
                "entombed",
            ),
            mock.call("All 2 links were ignored"),
            mock.call(),
            mock.call("Errors"),
            mock.call("------"),
            mock.call(
                ">> Immutable attribute not settable for "
                "/tmp/entomb_testing/fifo",
            ),
            mock.call(
                ">> Immutable attribute not settable for "
                "/tmp/entomb_testing/readable_by_root.txt",
            ),
            mock.call(),
            mock.call("\033[?25h", end=""),
        ]
        self.assertEqual(mocked_print.mock_calls, expected)

        # Confirm that no files were made immutable.
        for root_dir, _, filenames in os.walk(constants.DIRECTORY_PATH):
            for filename in filenames:
                file_path = os.path.join(root_dir, filename)
                with contextlib.suppress(subprocess.CalledProcessError):
                    self.assertFalse(helpers.file_is_immutable(file_path))

        # Test processing an empty directory.
        with mock.patch("builtins.print") as mocked_print:
            processing.process_objects(
                constants.EMPTY_SUBDIRECTORY_PATH,
                immutable=True,
                include_git=False,
                dry_run=False,
            )
        expected = [
            mock.call("\033[?25l", end=""),
            mock.call("Entomb objects"),
            mock.call(),
            mock.call("Progress"),
            mock.call("--------"),
            mock.call("Counting file paths: 0", end="\r"),
            mock.call("\033[K", end=""),
            mock.call("\033[K", end=""),
            mock.call(
                "████████████████████████████████████████",
                end="\r",
            ),
            mock.call(),
            mock.call(),
            mock.call("Summary"),
            mock.call("-------"),
            mock.call("No files were found"),
            mock.call(),
            mock.call("\033[?25h", end=""),
        ]
        self.assertEqual(mocked_print.mock_calls, expected)

        # Test processing a non-existent path.
        with self.assertRaises(AssertionError):
            processing.process_objects(
                constants.NON_EXISTENT_PATH,
                immutable=True,
                include_git=True,
                dry_run=False,
            )

        # Test a named pipe.
        with mock.patch("builtins.print") as mocked_print:
            processing.process_objects(
                constants.NAMED_PIPE_PATH,
                immutable=False,
                include_git=True,
                dry_run=False,
            )
        expected = [
            mock.call("\033[?25l", end=""),
            mock.call("Unset objects"),
            mock.call(),
            mock.call("Progress"),
            mock.call("--------"),
            mock.call("Counting file paths: 0", end="\r"),
            mock.call("\033[K", end=""),
            mock.call("\033[K", end=""),
            mock.call(
                "░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░  0.0%",
                end="\r",
            ),
            mock.call("\033[K", end=""),
            mock.call(
                "████████████████████████████████████████",
                end="\r",
            ),
            mock.call(),
            mock.call(),
            mock.call("Changes"),
            mock.call("-------"),
            mock.call("Unset 0 files"),
            mock.call(),
            mock.call("Summary"),
            mock.call("-------"),
            mock.call(
                "All 0 files for which immutability can be set are now unset",
            ),
            mock.call("All 0 links were ignored"),
            mock.call(),
            mock.call("Errors"),
            mock.call("------"),
            mock.call(
                ">> Immutable attribute not settable for "
                "/tmp/entomb_testing/fifo",
            ),
            mock.call(),
            mock.call("\033[?25h", end=""),
        ]
        self.assertEqual(mocked_print.mock_calls, expected)

        # Test a file which is readable only by root.
        with mock.patch("builtins.print") as mocked_print:
            processing.process_objects(
                constants.READABLE_BY_ROOT_FILE_PATH,
                immutable=True,
                include_git=False,
                dry_run=False,
            )
        expected = [
            mock.call("\033[?25l", end=""),
            mock.call("Entomb objects"),
            mock.call(),
            mock.call("Progress"),
            mock.call("--------"),
            mock.call("Counting file paths: 0", end="\r"),
            mock.call("\033[K", end=""),
            mock.call("\033[K", end=""),
            mock.call(
                "░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░  0.0%",
                end="\r",
            ),
            mock.call("\033[K", end=""),
            mock.call(
                "████████████████████████████████████████",
                end="\r",
            ),
            mock.call(),
            mock.call(),
            mock.call("Changes"),
            mock.call("-------"),
            mock.call("Entombed 0 files"),
            mock.call(),
            mock.call("Summary"),
            mock.call("-------"),
            mock.call(
                "All 0 files for which immutability can be set are now "
                "entombed",
            ),
            mock.call("All 0 links were ignored"),
            mock.call(),
            mock.call("Errors"),
            mock.call("------"),
            mock.call(
                ">> Immutable attribute not settable for "
                "/tmp/entomb_testing/readable_by_root.txt",
            ),
            mock.call(),
            mock.call("\033[?25h", end=""),
        ]
        self.assertEqual(mocked_print.mock_calls, expected)

    def test__print_errors(self):
        """Test the _print_errors function.

        """
        # Test with no errors.
        with mock.patch("builtins.print") as mocked_print:
            processing._print_errors([])
        expected = []
        self.assertEqual(mocked_print.mock_calls, expected)

        # Test with three errors.
        with mock.patch("builtins.print") as mocked_print:
            processing._print_errors([
                "ERROR: Message 1",
                "ERROR: Message 2",
                "ERROR: Message 3",
            ])
        expected = [
            mock.call("Errors"),
            mock.call("------"),
            mock.call(">> ERROR: Message 1"),
            mock.call(">> ERROR: Message 2"),
            mock.call(">> ERROR: Message 3"),
            mock.call(),
        ]
        self.assertEqual(mocked_print.mock_calls, expected)

        # Test with 12 errors.
        with mock.patch("builtins.print") as mocked_print:
            processing._print_errors([
                "ERROR: Message 1",
                "ERROR: Message 2",
                "ERROR: Message 3",
                "ERROR: Message 4",
                "ERROR: Message 5",
                "ERROR: Message 6",
                "ERROR: Message 7",
                "ERROR: Message 8",
                "ERROR: Message 9",
                "ERROR: Message 10",
                "ERROR: Message 11",
                "ERROR: Message 12",
            ])
        expected = [
            mock.call("Errors"),
            mock.call("------"),
            mock.call(">> ERROR: Message 1"),
            mock.call(">> ERROR: Message 2"),
            mock.call(">> ERROR: Message 3"),
            mock.call(">> ERROR: Message 4"),
            mock.call(">> ERROR: Message 5"),
            mock.call(">> ERROR: Message 6"),
            mock.call(">> ERROR: Message 7"),
            mock.call(">> ERROR: Message 8"),
            mock.call(">> ERROR: Message 9"),
            mock.call(">> ERROR: Message 10"),
            mock.call(">> Plus 2 more errors"),
            mock.call(),
        ]
        self.assertEqual(mocked_print.mock_calls, expected)

    def test__process_object(self):
        """Test the _process_object function.

        """
        # Test making a mutable file immutable, then return it to its original
        # state.
        processing._process_object(
            constants.MUTABLE_FILE_PATH,
            immutable=True,
            dry_run=False,
        )
        self.assertTrue(helpers.file_is_immutable(constants.MUTABLE_FILE_PATH))
        processing._process_object(
            constants.MUTABLE_FILE_PATH,
            immutable=False,
            dry_run=False,
        )

        # Test making a mutable file mutable.
        processing._process_object(
            constants.MUTABLE_FILE_PATH,
            immutable=False,
            dry_run=False,
        )
        self.assertFalse(
            helpers.file_is_immutable(constants.MUTABLE_FILE_PATH),
        )

        # Test making an immutable file mutable, then return it to its original
        # state.
        processing._process_object(
            constants.IMMUTABLE_FILE_PATH,
            immutable=False,
            dry_run=False,
        )
        self.assertFalse(
            helpers.file_is_immutable(constants.IMMUTABLE_FILE_PATH),
        )
        processing._process_object(
            constants.IMMUTABLE_FILE_PATH,
            immutable=True,
            dry_run=False,
        )

        # Test making an immutable file immutable.
        processing._process_object(
            constants.IMMUTABLE_FILE_PATH,
            immutable=True,
            dry_run=False,
        )
        self.assertTrue(
            helpers.file_is_immutable(constants.IMMUTABLE_FILE_PATH),
        )

        # Test making an immutable file mutable as a dry run.
        processing._process_object(
            constants.IMMUTABLE_FILE_PATH,
            immutable=False,
            dry_run=True,
        )
        self.assertTrue(
            helpers.file_is_immutable(constants.IMMUTABLE_FILE_PATH),
        )

        # Test making a link immutable.
        with self.assertRaises(AssertionError):
            processing._process_object(
                constants.LINK_PATH,
                immutable=True,
                dry_run=False,
            )

        # Test making a directory mutable.
        with self.assertRaises(AssertionError):
            processing._process_object(
                constants.DIRECTORY_PATH,
                immutable=False,
                dry_run=False,
            )

        # Test making a non-existent path mutable.
        with self.assertRaises(AssertionError):
            processing._process_object(
                constants.NON_EXISTENT_PATH,
                immutable=False,
                dry_run=False,
            )

    def test__set_attribute(self):
        """Test the _set_attribute function.

        """
        # Test making an immutable file mutable.
        processing._set_attribute("-i", constants.IMMUTABLE_FILE_PATH)
        self.assertFalse(
            helpers.file_is_immutable(constants.IMMUTABLE_FILE_PATH),
        )

        # Test making a mutable file immutable.
        processing._set_attribute("+i", constants.IMMUTABLE_FILE_PATH)
        self.assertTrue(
            helpers.file_is_immutable(constants.IMMUTABLE_FILE_PATH),
        )

        # Test making a non-existent path to immutable.
        with self.assertRaises(exceptions.SetAttributeError):
            processing._set_attribute("+i", constants.NON_EXISTENT_PATH)

    def tearDown(self):
        """Delete temporary directories and files.

        """
        helpers.tear_down()



# Command line arguments.
DRY_RUN_ARG = "--dry-run"
DRY_RUN_SHORT_ARG = "-d"
INCLUDE_GIT_ARG = "--include-git"
INCLUDE_GIT_SHORT_ARG = "-g"
LIST_IMMUTABLE_ARG = "--list-immutable"
LIST_MUTABLE_ARG = "--list-mutable"
REPORT_ARG = "--report"
REPORT_SHORT_ARG = "-r"
UNSET_ARG = "--unset"
UNSET_SHORT_ARG = "-u"
VERSION_ARG = "--version"
VERSION_SHORT_ARG = "-v"


# Formatting
TABLE_WIDTH = 40


# Path types.
DIRECTORY = "directory"
FILE = "file"
LINK = "link"



import setuptools

import entomb


with open("README.rst", "r") as f:
    LONG_DESCRIPTION = f.read()

NAME = entomb.__title__.lower()

setuptools.setup(
    name=NAME,
    version=entomb.__version__,
    description=entomb.__description__,
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/x-rst",
    author=entomb.__author__,
    author_email=entomb.__author_email__,
    license=entomb.__licence__,
    url=entomb.__url__,
    packages=setuptools.find_packages(),
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "License :: OSI Approved :: MIT License",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Topic :: Utilities",
    ],
    keywords="immutable",
    python_requires=">=3.5",
    entry_points={
        "console_scripts": [
            "{}=entomb.core:run".format(NAME),
        ],
    },
)



import contextlib
import os
import shutil
import subprocess

from tests import constants


def file_is_immutable(path):
    """Whether a file has the immutable attribute set.

    Parameters
    ----------
    path : str
        An absolute path to a file.

    Returns
    -------
    bool
        True if the file's immmutable attribute is set, False if it is not.

    Raises
    ------
    CalledProcessError
        If the exit status of the chattr command is non-zero.

    """
    # Run the lsattr command.
    lsattr_result = subprocess.run(
        ["lsattr", path],
        check=True,
        stderr=subprocess.DEVNULL,
        stdout=subprocess.PIPE,
        universal_newlines=True,
    )

    # Extract the immutable attribute from the command output.
    attributes = lsattr_result.stdout.split()[0]
    immutable_flag = list(attributes)[4]

    return immutable_flag == "i"


def set_file_immutable_attribute(path, immutable):
    """Set or unset the immutable attribute for a file.

    Parameters
    ----------
    path : str
        The absolute path of a file.
    immutable: bool
        Set immutable attribute if True, unset immutable attribute if False.

    Returns
    -------
    None

    Raises
    ------
    CalledProcessError
        If the exit status of the chattr command is non-zero.

    """
    operation = "+i" if immutable else "-i"

    subprocess.run(
        ["sudo", "chattr", operation, path],
        check=True,
        stderr=subprocess.DEVNULL,
    )


def set_up():
    """Create temporary directories and files.

    Returns
    -------
    None

    """
    # Ensure that tests start with a clean slate.
    tear_down()

    # Create testing directories.
    os.makedirs(constants.EMPTY_SUBDIRECTORY_PATH)
    os.makedirs(constants.GIT_SUBDIRECTORY_PATH)
    os.makedirs(constants.SUBDIRECTORY_PATH)

    # Create testing files.
    open(constants.GIT_DIRECTORY_MUTABLE_FILE_PATH, "x").close()
    open(constants.GIT_SUBDIRECTORY_MUTABLE_FILE_PATH, "x").close()
    open(constants.IMMUTABLE_FILE_PATH, "x").close()
    open(constants.MUTABLE_FILE_PATH, "x").close()
    open(constants.SUBDIRECTORY_IMMUTABLE_FILE_PATH, "x").close()
    open(constants.SUBDIRECTORY_MUTABLE_FILE_PATH, "x").close()
    open(constants.READABLE_BY_ROOT_FILE_PATH, "x").close()

    # Create testing named pipe.
    os.mkfifo(constants.NAMED_PIPE_PATH)

    # Create testing links.
    os.symlink(constants.MUTABLE_FILE_PATH, constants.LINK_PATH)
    os.symlink(
        constants.SUBDIRECTORY_MUTABLE_FILE_PATH,
        constants.SUBDIRECTORY_LINK_PATH,
    )

    # Set immutability for some testing files.
    set_file_immutable_attribute(constants.IMMUTABLE_FILE_PATH, immutable=True)
    set_file_immutable_attribute(
        constants.SUBDIRECTORY_IMMUTABLE_FILE_PATH,
        immutable=True,
    )

    # Set ownership and permissions of the file which is readable only by root.
    os.chmod(constants.READABLE_BY_ROOT_FILE_PATH, 0o400)
    subprocess.run(
        ["sudo", "chown", "root:root", constants.READABLE_BY_ROOT_FILE_PATH],
        check=True,
    )


def tear_down():
    """Delete temporary directories and files.

    Returns
    -------
    None

    """
    # Ensure all testing files are mutable, or they won't able to be deleted.
    for root_dir, _, filenames in os.walk(constants.DIRECTORY_PATH):
        for filename in filenames:
            file_path = os.path.join(root_dir, filename)
            with contextlib.suppress(subprocess.CalledProcessError):
                set_file_immutable_attribute(file_path, immutable=False)

    # Remove the testing directory.
    try:
        shutil.rmtree(constants.DIRECTORY_PATH)
    except FileNotFoundError:
        pass



import unittest
import unittest.mock as mock

import entomb.reporting as reporting
from tests import (
    constants,
    helpers,
)


class TestReporting(unittest.TestCase):
    """Tests for the reporting module.

    """

    def setUp(self):
        """Create temporary directories and files.

        """
        helpers.set_up()

    def test_print_report(self):
        """Test the print_report function.

        """
        # Test a link.
        with mock.patch("builtins.print") as mocked_print:
            reporting.produce_report(constants.LINK_PATH, include_git=True)
        expected = [
            mock.call("\033[?25l", end=""),
            mock.call("Produce report"),
            mock.call(),
            mock.call("Report"),
            mock.call("------"),
            mock.call("A link has no immutable attribute"),
            mock.call(),
            mock.call("\033[?25h", end=""),
        ]
        self.assertEqual(mocked_print.mock_calls, expected)

        # Test an immutable file.
        with mock.patch("builtins.print") as mocked_print:
            reporting.produce_report(
                constants.IMMUTABLE_FILE_PATH,
                include_git=False,
            )
        expected = [
            mock.call("\033[?25l", end=""),
            mock.call("Produce report"),
            mock.call(),
            mock.call("Report"),
            mock.call("------"),
            mock.call("File is immutable"),
            mock.call(),
            mock.call("\033[?25h", end=""),
        ]
        self.assertEqual(mocked_print.mock_calls, expected)

        # Test a mutable file.
        with mock.patch("builtins.print") as mocked_print:
            reporting.produce_report(
                constants.MUTABLE_FILE_PATH,
                include_git=False,
            )
        expected = [
            mock.call("\033[?25l", end=""),
            mock.call("Produce report"),
            mock.call(),
            mock.call("Report"),
            mock.call("------"),
            mock.call("File is mutable"),
            mock.call(),
            mock.call("\033[?25h", end=""),
        ]
        self.assertEqual(mocked_print.mock_calls, expected)

        # Test a named pipe.
        with mock.patch("builtins.print") as mocked_print:
            reporting.produce_report(
                constants.NAMED_PIPE_PATH,
                include_git=False,
            )
        expected = [
            mock.call("\033[?25l", end=""),
            mock.call("Produce report"),
            mock.call(),
            mock.call("Report"),
            mock.call("------"),
            mock.call("Immutable attribute could not be accessed"),
            mock.call(),
            mock.call("\033[?25h", end=""),
        ]
        self.assertEqual(mocked_print.mock_calls, expected)

        # Test a file which is readable only by root.
        with mock.patch("builtins.print") as mocked_print:
            reporting.produce_report(
                constants.READABLE_BY_ROOT_FILE_PATH,
                include_git=True,
            )
        expected = [
            mock.call("\033[?25l", end=""),
            mock.call("Produce report"),
            mock.call(),
            mock.call("Report"),
            mock.call("------"),
            mock.call("Immutable attribute could not be accessed"),
            mock.call(),
            mock.call("\033[?25h", end=""),
        ]
        self.assertEqual(mocked_print.mock_calls, expected)

        # Test a directory including git.
        with mock.patch("builtins.print") as mocked_print:
            reporting.produce_report(
                constants.DIRECTORY_PATH,
                include_git=True,
            )
        expected = [
            mock.call("\033[?25l", end=""),
            mock.call("Produce report"),
            mock.call(),
            mock.call("Progress"),
            mock.call("--------"),
            mock.call("Counting file paths: 0", end="\r"),
            mock.call("\033[K", end=""),
            mock.call("\033[K", end=""),
            mock.call(
                "░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░  0.0%",
                end="\r",
            ),
            mock.call("\033[K", end=""),
            mock.call(
                "████████████████████████████████████████",
                end="\r",
            ),
            mock.call(),
            mock.call(),
            mock.call("Report"),
            mock.call("----------------------------------------"),
            mock.call("Immutable files", "                       2"),
            mock.call("----------------------------------------"),
            mock.call("Mutable files", "                         4"),
            mock.call("----------------------------------------"),
            mock.call("Inaccessible files", "                    2"),
            mock.call("----------------------------------------"),
            mock.call("All files", "                             8"),
            mock.call("----------------------------------------"),
            mock.call("Entombed", "                            25%"),
            mock.call("----------------------------------------"),
            mock.call("Links", "                                 2"),
            mock.call("----------------------------------------"),
            mock.call("Sub-directories", "                       4"),
            mock.call("----------------------------------------"),
            mock.call(),
            mock.call("\033[?25h", end=""),
        ]
        self.assertEqual(mocked_print.mock_calls, expected)

        # Test a directory excluding git.
        with mock.patch("builtins.print") as mocked_print:
            reporting.produce_report(
                constants.DIRECTORY_PATH,
                include_git=False,
            )
        expected = [
            mock.call("\033[?25l", end=""),
            mock.call("Produce report"),
            mock.call(),
            mock.call("Progress"),
            mock.call("--------"),
            mock.call("Counting file paths: 0", end="\r"),
            mock.call("\033[K", end=""),
            mock.call("\033[K", end=""),
            mock.call(
                "░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░  0.0%",
                end="\r",
            ),
            mock.call("\033[K", end=""),
            mock.call(
                "████████████████████████████████████████",
                end="\r",
            ),
            mock.call(),
            mock.call(),
            mock.call("Report"),
            mock.call("----------------------------------------"),
            mock.call("Immutable files", "                       2"),
            mock.call("----------------------------------------"),
            mock.call("Mutable files", "                         2"),
            mock.call("----------------------------------------"),
            mock.call("Inaccessible files", "                    2"),
            mock.call("----------------------------------------"),
            mock.call("All files", "                             6"),
            mock.call("----------------------------------------"),
            mock.call("Entombed", "                            33%"),
            mock.call("----------------------------------------"),
            mock.call("Links", "                                 2"),
            mock.call("----------------------------------------"),
            mock.call("Sub-directories", "                       2"),
            mock.call("----------------------------------------"),
            mock.call(),
            mock.call("\033[?25h", end=""),
        ]
        self.assertEqual(mocked_print.mock_calls, expected)

        # Test an empty directory.
        with mock.patch("builtins.print") as mocked_print:
            reporting.produce_report(
                constants.EMPTY_SUBDIRECTORY_PATH,
                include_git=False,
            )
        expected = [
            mock.call("\033[?25l", end=""),
            mock.call("Produce report"),
            mock.call(),
            mock.call("Progress"),
            mock.call("--------"),
            mock.call("Counting file paths: 0", end="\r"),
            mock.call("\033[K", end=""),
            mock.call("\033[K", end=""),
            mock.call(
                "████████████████████████████████████████",
                end="\r",
            ),
            mock.call(),
            mock.call(),
            mock.call("Report"),
            mock.call("----------------------------------------"),
            mock.call("Immutable files", "                       0"),
            mock.call("----------------------------------------"),
            mock.call("Mutable files", "                         0"),
            mock.call("----------------------------------------"),
            mock.call("All files", "                             0"),
            mock.call("----------------------------------------"),
            mock.call("Entombed", "                            n/a"),
            mock.call("----------------------------------------"),
            mock.call("Links", "                                 0"),
            mock.call("----------------------------------------"),
            mock.call("Sub-directories", "                       0"),
            mock.call("----------------------------------------"),
            mock.call(),
            mock.call("\033[?25h", end=""),
        ]
        self.assertEqual(mocked_print.mock_calls, expected)

        # Test a non-existent path.
        with self.assertRaises(AssertionError):
            reporting.produce_report(
                constants.NON_EXISTENT_PATH,
                include_git=False,
            )

    def test__print_abbreviated_report(self):
        """Test the _print_abbreviated_report function.

        """
        # Test an immutable file.
        with mock.patch("builtins.print") as mocked_print:
            reporting._print_abbreviated_report(constants.IMMUTABLE_FILE_PATH)
        expected = [
            mock.call("Report"),
            mock.call("------"),
            mock.call("File is immutable"),
            mock.call(),
        ]
        self.assertEqual(mocked_print.mock_calls, expected)

        # Test a mutable file.
        with mock.patch("builtins.print") as mocked_print:
            reporting._print_abbreviated_report(constants.MUTABLE_FILE_PATH)
        expected = [
            mock.call("Report"),
            mock.call("------"),
            mock.call("File is mutable"),
            mock.call(),
        ]
        self.assertEqual(mocked_print.mock_calls, expected)

        # Test a link.
        with mock.patch("builtins.print") as mocked_print:
            reporting._print_abbreviated_report(constants.LINK_PATH)
        expected = [
            mock.call("Report"),
            mock.call("------"),
            mock.call("A link has no immutable attribute"),
            mock.call(),
        ]
        self.assertEqual(mocked_print.mock_calls, expected)

        # Test a directory.
        with self.assertRaises(AssertionError):
            reporting._print_abbreviated_report(constants.DIRECTORY_PATH)

        # Test a non-existent path.
        with self.assertRaises(AssertionError):
            reporting._print_abbreviated_report(constants.NON_EXISTENT_PATH)

        # Test a named pipe.
        with mock.patch("builtins.print") as mocked_print:
            reporting._print_abbreviated_report(constants.NAMED_PIPE_PATH)
        expected = [
            mock.call("Report"),
            mock.call("------"),
            mock.call("Immutable attribute could not be accessed"),
            mock.call(),
        ]
        self.assertEqual(mocked_print.mock_calls, expected)

        # Test a file which is readable only by root.
        with mock.patch("builtins.print") as mocked_print:
            reporting._print_abbreviated_report(
                constants.READABLE_BY_ROOT_FILE_PATH,
            )
        expected = [
            mock.call("Report"),
            mock.call("------"),
            mock.call("Immutable attribute could not be accessed"),
            mock.call(),
        ]
        self.assertEqual(mocked_print.mock_calls, expected)

    def test__print_full_report(self):
        """Test the _print_full_report function.

        """
        # Test with a non-zero entombed percentage.
        with mock.patch("builtins.print") as mocked_print:
            reporting._print_full_report(
                directory_count=6,
                link_count=5,
                immutable_file_count=8,
                inaccessible_file_count=3,
                mutable_file_count=27,
            )
        expected = [
            mock.call("Report"),
            mock.call("----------------------------------------"),
            mock.call("Immutable files", "                       8"),
            mock.call("----------------------------------------"),
            mock.call("Mutable files", "                        27"),
            mock.call("----------------------------------------"),
            mock.call("Inaccessible files", "                    3"),
            mock.call("----------------------------------------"),
            mock.call("All files", "                            38"),
            mock.call("----------------------------------------"),
            mock.call("Entombed", "                            21%"),
            mock.call("----------------------------------------"),
            mock.call("Links", "                                 5"),
            mock.call("----------------------------------------"),
            mock.call("Sub-directories", "                       5"),
            mock.call("----------------------------------------"),
            mock.call(),
        ]
        self.assertEqual(mocked_print.mock_calls, expected)

        # Test with an n/a entombed percentage.
        with mock.patch("builtins.print") as mocked_print:
            reporting._print_full_report(
                directory_count=3,
                link_count=0,
                immutable_file_count=0,
                inaccessible_file_count=0,
                mutable_file_count=0,
            )
        expected = [
            mock.call("Report"),
            mock.call("----------------------------------------"),
            mock.call("Immutable files", "                       0"),
            mock.call("----------------------------------------"),
            mock.call("Mutable files", "                         0"),
            mock.call("----------------------------------------"),
            mock.call("All files", "                             0"),
            mock.call("----------------------------------------"),
            mock.call("Entombed", "                            n/a"),
            mock.call("----------------------------------------"),
            mock.call("Links", "                                 0"),
            mock.call("----------------------------------------"),
            mock.call("Sub-directories", "                       2"),
            mock.call("----------------------------------------"),
            mock.call(),
        ]
        self.assertEqual(mocked_print.mock_calls, expected)

    def tearDown(self):
        """Delete temporary directories and files.

        """
        helpers.tear_down()



import datetime
import unittest
import unittest.mock as mock

import entomb.exceptions as exceptions
import entomb.utilities as utilities
from tests import (
    constants,
    helpers,
)


class TestUtilities(unittest.TestCase):
    """Tests for the utilities module.

    """

    def setUp(self):
        """Create temporary directories and files.

        """
        helpers.set_up()

    def test_clear_line(self):
        """Test the clear_line function.

        """
        with mock.patch("builtins.print") as mocked_print:
            utilities.clear_line()
        expected = [mock.call("\033[K", end="")]
        self.assertEqual(mocked_print.mock_calls, expected)

    def test_count_file_paths(self):
        """Test the count_file_paths function.

        """
        # Test a directory excluding git files with a print frequency of 1.
        with mock.patch("builtins.print") as mocked_print:
            actual_count = utilities.count_file_paths(
                constants.DIRECTORY_PATH,
                include_git=False,
                print_frequency=1,
            )
            expected_count = 8
            self.assertEqual(actual_count, expected_count)

        expected_output = [
            mock.call("Counting file paths: 0", end="\r"),
            mock.call("Counting file paths: 1", end="\r"),
            mock.call("Counting file paths: 2", end="\r"),
            mock.call("Counting file paths: 3", end="\r"),
            mock.call("Counting file paths: 4", end="\r"),
            mock.call("Counting file paths: 5", end="\r"),
            mock.call("Counting file paths: 6", end="\r"),
            mock.call("Counting file paths: 7", end="\r"),
            mock.call("Counting file paths: 8", end="\r"),
            mock.call("\033[K", end=""),
        ]
        self.assertEqual(mocked_print.mock_calls, expected_output)

        # Test a directory including git files with a print frequency of 2.
        with mock.patch("builtins.print") as mocked_print:
            actual_count = utilities.count_file_paths(
                constants.DIRECTORY_PATH,
                include_git=True,
                print_frequency=2,
            )
            expected_count = 10
            self.assertEqual(actual_count, expected_count)

        expected_output = [
            mock.call("Counting file paths: 0", end="\r"),
            mock.call("Counting file paths: 2", end="\r"),
            mock.call("Counting file paths: 4", end="\r"),
            mock.call("Counting file paths: 6", end="\r"),
            mock.call("Counting file paths: 8", end="\r"),
            mock.call("Counting file paths: 10", end="\r"),
            mock.call("\033[K", end=""),
        ]
        self.assertEqual(mocked_print.mock_calls, expected_output)

        # Test an empty directory.
        with mock.patch("builtins.print") as mocked_print:
            actual_count = utilities.count_file_paths(
                constants.EMPTY_SUBDIRECTORY_PATH,
                include_git=False,
                print_frequency=1,
            )
            expected_count = 0
            self.assertEqual(actual_count, expected_count)

        expected_output = [
            mock.call("Counting file paths: 0", end="\r"),
            mock.call("\033[K", end=""),
        ]
        self.assertEqual(mocked_print.mock_calls, expected_output)

    def test_file_is_immutable(self):
        """Test the file_is_immutable function.

        """
        # Test an immutable file path.
        self.assertTrue(
            utilities.file_is_immutable(constants.IMMUTABLE_FILE_PATH),
        )

        # Test a mutable file path.
        self.assertFalse(
            utilities.file_is_immutable(constants.MUTABLE_FILE_PATH),
        )

        # Test a directory path.
        with self.assertRaises(AssertionError):
            utilities.file_is_immutable(constants.DIRECTORY_PATH)

        # Test a link path.
        with self.assertRaises(AssertionError):
            utilities.file_is_immutable(constants.LINK_PATH)

        # Test a path which does not exist.
        with self.assertRaises(AssertionError):
            utilities.file_is_immutable(constants.NON_EXISTENT_PATH)

        # Test a string which can't be parsed as a path.
        with self.assertRaises(AssertionError):
            utilities.file_is_immutable(constants.NON_PATH_STRING)

        # Test a named pipe.
        with self.assertRaises(exceptions.GetAttributeError):
            utilities.file_is_immutable(constants.NAMED_PIPE_PATH)

        # Test a file which is readable only by root.
        with self.assertRaises(exceptions.GetAttributeError):
            utilities.file_is_immutable(constants.READABLE_BY_ROOT_FILE_PATH)

    def test_file_paths(self):
        """Test the file_paths function.

        """
        # Test a directory excluding git files.
        actual = utilities.file_paths(
            constants.DIRECTORY_PATH,
            include_git=False,
        )
        expected = [
            constants.IMMUTABLE_FILE_PATH,
            constants.LINK_PATH,
            constants.MUTABLE_FILE_PATH,
            constants.NAMED_PIPE_PATH,
            constants.READABLE_BY_ROOT_FILE_PATH,
            constants.SUBDIRECTORY_IMMUTABLE_FILE_PATH,
            constants.SUBDIRECTORY_LINK_PATH,
            constants.SUBDIRECTORY_MUTABLE_FILE_PATH,
        ]
        self.assertEqual(sorted(actual), sorted(expected))

        # Test a directory including git files.
        actual = utilities.file_paths(
            constants.DIRECTORY_PATH,
            include_git=True,
        )
        expected = [
            constants.GIT_DIRECTORY_MUTABLE_FILE_PATH,
            constants.GIT_SUBDIRECTORY_MUTABLE_FILE_PATH,
            constants.IMMUTABLE_FILE_PATH,
            constants.LINK_PATH,
            constants.MUTABLE_FILE_PATH,
            constants.NAMED_PIPE_PATH,
            constants.READABLE_BY_ROOT_FILE_PATH,
            constants.SUBDIRECTORY_IMMUTABLE_FILE_PATH,
            constants.SUBDIRECTORY_LINK_PATH,
            constants.SUBDIRECTORY_MUTABLE_FILE_PATH,
        ]
        self.assertEqual(sorted(actual), sorted(expected))

        # Test a file.
        actual = utilities.file_paths(
            constants.IMMUTABLE_FILE_PATH,
            include_git=False,
        )
        expected = [constants.IMMUTABLE_FILE_PATH]
        self.assertEqual(sorted(actual), sorted(expected))

        # Test a link.
        actual = utilities.file_paths(constants.LINK_PATH, include_git=False)
        expected = [constants.LINK_PATH]
        self.assertEqual(sorted(actual), sorted(expected))

        # Test a directory which does not exist.
        with self.assertRaises(AssertionError):
            paths = utilities.file_paths(
                constants.NON_PATH_STRING,
                include_git=True,
            )
            # Ths exception will only be raised once the generator is iterated.
            next(paths)

        # Test a named pipe.
        actual = utilities.file_paths(
            constants.NAMED_PIPE_PATH,
            include_git=False,
        )
        expected = [constants.NAMED_PIPE_PATH]
        self.assertEqual(sorted(actual), sorted(expected))

        # Test a file which is readable only by root.
        actual = utilities.file_paths(
            constants.READABLE_BY_ROOT_FILE_PATH,
            include_git=True,
        )
        expected = [constants.READABLE_BY_ROOT_FILE_PATH]
        self.assertEqual(sorted(actual), sorted(expected))

    def test_hide_cursor(self):
        """Test the hide_cursor function.

        """
        with mock.patch("builtins.print") as mocked_print:
            # The function is used as a decorator, but is easier to test as a
            # context manager.
            with utilities.hide_cursor():
                print("testing")
        expected = [
            mock.call("\033[?25l", end=""),
            mock.call("testing"),
            mock.call("\033[?25h", end=""),
        ]
        self.assertEqual(mocked_print.mock_calls, expected)

    def test_print_header(self):
        """Test the print_header function.

        """
        with mock.patch("builtins.print") as mocked_print:
            utilities.print_header("Header")
        expected = [
            mock.call("Header"),
            mock.call("------"),
        ]
        self.assertEqual(mocked_print.mock_calls, expected)

    def test_print_progress_bar(self):
        """Test the progress_bar function.

        """
        start_time = datetime.datetime.now()

        # Test progress of exactly 0%.
        with mock.patch("builtins.print") as mocked_print:
            with mock.patch("datetime.datetime") as mocked_datetime:
                now = start_time + datetime.timedelta(microseconds=10)
                mocked_datetime.now.return_value = now
                utilities.print_progress_bar(
                    start_time=start_time,
                    count=0,
                    total=10000,
                    print_frequency=100,
                )
        expected_progress_bar = (
            "░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░  0.0%"
        )
        expected = [
            mock.call("\033[K", end=""),
            mock.call(expected_progress_bar, end="\r"),
            ]
        self.assertEqual(mocked_print.mock_calls, expected)

        # Test progress of exactly 50%.
        with mock.patch("builtins.print") as mocked_print:
            with mock.patch("datetime.datetime") as mocked_datetime:
                now = start_time + datetime.timedelta(seconds=5)
                mocked_datetime.now.return_value = now
                utilities.print_progress_bar(
                    start_time=start_time,
                    count=5000,
                    total=10000,
                    print_frequency=100,
                )
        expected_progress_bar = (
            "████████████████████░░░░░░░░░░░░░░░░░░░░  50.0%  |  5s to go"
        )
        expected = [
            mock.call("\033[K", end=""),
            mock.call(expected_progress_bar, end="\r"),
            ]
        self.assertEqual(mocked_print.mock_calls, expected)

        # Test progress of exactly 100%.
        with mock.patch("builtins.print") as mocked_print:
            with mock.patch("datetime.datetime") as mocked_datetime:
                now = start_time + datetime.timedelta(seconds=10)
                mocked_datetime.now.return_value = now
                utilities.print_progress_bar(
                    start_time=start_time,
                    count=12345,
                    total=12345,
                    print_frequency=100,
                )
        expected_progress_bar = "████████████████████████████████████████"
        expected = [
            mock.call("\033[K", end=""),
            mock.call(expected_progress_bar, end="\r"),
            ]
        self.assertEqual(mocked_print.mock_calls, expected)

        # Test progress rounding down to 99.9% rather than up to 100%.
        with mock.patch("builtins.print") as mocked_print:
            with mock.patch("datetime.datetime") as mocked_datetime:
                now = start_time + datetime.timedelta(seconds=10)
                mocked_datetime.now.return_value = now
                utilities.print_progress_bar(
                    start_time=start_time,
                    count=9999,
                    total=10000,
                    print_frequency=1,
                )
        expected_progress_bar = (
            "███████████████████████████████████████░  99.9%  |  0s to go"
        )
        expected = [
            mock.call("\033[K", end=""),
            mock.call(expected_progress_bar, end="\r"),
            ]
        self.assertEqual(mocked_print.mock_calls, expected)

        # Test printing nothing when frequency condition is not met.
        with mock.patch("builtins.print") as mocked_print:
            with mock.patch("datetime.datetime") as mocked_datetime:
                now = start_time + datetime.timedelta(seconds=1)
                mocked_datetime.now.return_value = now
                utilities.print_progress_bar(
                    start_time=start_time,
                    count=1001,
                    total=10000,
                    print_frequency=100,
                )
        expected = []
        self.assertEqual(mocked_print.mock_calls, expected)

        # Test for count of 0 and total of 0.
        with mock.patch("builtins.print") as mocked_print:
            with mock.patch("datetime.datetime") as mocked_datetime:
                now = start_time + datetime.timedelta(milliseconds=10)
                mocked_datetime.now.return_value = now
                utilities.print_progress_bar(
                    start_time=start_time,
                    count=0,
                    total=0,
                    print_frequency=1,
                )
        expected_progress_bar = "████████████████████████████████████████"
        expected = [
            mock.call("\033[K", end=""),
            mock.call(expected_progress_bar, end="\r"),
        ]
        self.assertEqual(mocked_print.mock_calls, expected)

        # Test for an operation which takes several hours.
        with mock.patch("builtins.print") as mocked_print:
            with mock.patch("datetime.datetime") as mocked_datetime:
                now = start_time + datetime.timedelta(hours=1, minutes=23)
                mocked_datetime.now.return_value = now
                utilities.print_progress_bar(
                    start_time=start_time,
                    count=5840300,
                    total=21735498,
                    print_frequency=100,
                )
        expected_progress_bar = (
            "██████████░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░  "
            "26.8%  |  3h 45m 54s to go"
        )
        expected = [
            mock.call("\033[K", end=""),
            mock.call(expected_progress_bar, end="\r"),
            ]
        self.assertEqual(mocked_print.mock_calls, expected)

    def test__add_percentage_to_progress_bar(self):
        """Test the _add_percentage_to_progress_bar function.

        """
        # Because _add_percentage_to_progress_bar() contributes to
        # print_progress_bar() it is not tested in isolation, but is tested as
        # part of print_progress_bar() by test_print_progress_bar().

    def test__add_time_to_progress_bar(self):
        """Test the _add_time_to_progress_bar function.

        """
        # Because _add_time_to_progress_bar() contributes to
        # print_progress_bar() it is not tested in isolation, but is tested as
        # part of print_progress_bar() by test_print_progress_bar().

    def test__build_progress_bar(self):
        """Test the _build_progress_bar function.

        """
        # Because _build_progress_bar() contributes to print_progress_bar() it
        # is not tested in isolation, but is tested as part of
        # print_progress_bar() by test_print_progress_bar().

    def test__get_immutable_flag(self):
        """Test the _get_immutable_flag function.

        """
        # Test an immutable file.
        actual = utilities._get_immutable_flag(constants.IMMUTABLE_FILE_PATH)
        expected = "i"
        self.assertEqual(actual, expected)

        # Test a mutable file.
        actual = utilities._get_immutable_flag(constants.MUTABLE_FILE_PATH)
        expected = "-"
        self.assertEqual(actual, expected)

        # Test a link.
        with self.assertRaises(exceptions.GetAttributeError):
            actual = utilities._get_immutable_flag(constants.LINK_PATH)

    def test__readable_duration(self):
        """Test the _readable_duration function.

        """
        self.assertEqual(utilities._readable_duration(0), "0s")
        self.assertEqual(utilities._readable_duration(5), "5s")
        self.assertEqual(utilities._readable_duration(59.1), "59s")
        self.assertEqual(utilities._readable_duration(59.9), "1m 00s")
        self.assertEqual(utilities._readable_duration(305), "5m 05s")
        self.assertEqual(utilities._readable_duration(600.1), "10m 00s")
        self.assertEqual(utilities._readable_duration(3600), "1h 00m 00s")
        self.assertEqual(
            utilities._readable_duration(3600000),
            "1,000h 00m 00s",
        )

    def tearDown(self):
        """Delete temporary directories and files.

        """
        helpers.tear_down()
